@@ -17,6 +17,7 @@
 
 import contextlib
 import functools
+import itertools
 import os
 import sys
 import threading
@@ -147,9 +148,15 @@ class Config:
   def parse_flags_with_absl(self):
     global already_configured_with_absl
     if not already_configured_with_absl:
+      # Extract just the --jax... flags (before the first --) from argv. In some
+      # environments (e.g. ipython/colab) argv might be a mess of things
+      # parseable by absl and other junk.
+      jax_argv = itertools.takewhile(lambda a: a != '--', sys.argv)
+      jax_argv = ['', *(a for a in jax_argv if a.startswith('--jax'))]
+
       import absl.flags
       self.config_with_absl()
-      absl.flags.FLAGS(sys.argv, known_only=True)
+      absl.flags.FLAGS(jax_argv, known_only=True)
       self.complete_absl_config(absl.flags)
       already_configured_with_absl = True
 
@@ -228,7 +235,7 @@ class Config:
       return val if val is not unset else self._read(name)
     setattr(Config, name, property(get_state))
 
-    return _BoolStateContextManager(name, help, update_thread_local_hook)
+    return _StateContextManager(name, help, update_thread_local_hook)
 
   def define_enum_state(
       self, name: str, enum_values: List[str], default: Optional[str],
@@ -263,27 +270,55 @@ class Config:
       return val if val is not unset else self._read(name)
     setattr(Config, name, property(get_state))
 
-    @contextlib.contextmanager
-    def set_state(new_val: Optional[str]):
+    def validate(new_val):
       if (new_val is not None and
           (type(new_val) is not str or new_val not in enum_values)):
         raise ValueError(f"new enum value must be None or in {enum_values}, "
                          f"got {new_val} of type {type(new_val)}.")
-      prev_val = getattr(_thread_local_state, name, unset)
-      setattr(_thread_local_state, name, new_val)
-      if update_thread_local_hook: update_thread_local_hook(new_val)
-      try:
-        yield
-      finally:
-        if prev_val is unset:
-          delattr(_thread_local_state, name)
-          if update_thread_local_hook: update_thread_local_hook(None)
-        else:
-          setattr(_thread_local_state, name, prev_val)
-          if update_thread_local_hook: update_thread_local_hook(prev_val)
-    set_state.__name__ = name[4:] if name.startswith('jax_') else name
-    set_state.__doc__ = f"Context manager for `{name}` config option.\n\n{help}"
-    return set_state
+
+    return _StateContextManager(name, help, update_thread_local_hook, validate)
+
+  def define_string_state(
+      self, name: str, default: Optional[str], help: str,
+      update_global_hook: Optional[Callable[[str], None]] = None,
+      update_thread_local_hook: Optional[Callable[[Optional[str]], None]] = None):
+    """Set up thread-local state and return a contextmanager for managing it.
+
+    See docstring for ``define_bool_state``.
+
+    Args:
+      name: string, converted to lowercase to define the name of the config
+        option (and absl flag). It is converted to uppercase to define the
+        corresponding shell environment variable.
+      default: string, a default value for the option.
+      help: string, used to populate the flag help information as well as the
+        docstring of the returned context manager.
+      update_global_hook: an optional callback that is called with the updated
+        value of the global state when it is altered or set initially.
+      update_thread_local_hook: an optional callback that is called with the
+        updated value of the thread-local state when it is altered or set
+        initially.
+
+    Returns:
+      A contextmanager to control the thread-local state value.
+    """
+    name = name.lower()
+    default = os.getenv(name.upper(), default)
+    self.DEFINE_string(name, default, help=help,
+                       update_hook=update_global_hook)
+    self._contextmanager_flags.add(name)
+
+    def get_state(self):
+      val = getattr(_thread_local_state, name, unset)
+      return val if val is not unset else self._read(name)
+    setattr(Config, name, property(get_state))
+
+    def validate(new_val):
+      if new_val is not None and not isinstance(new_val, str):
+        raise ValueError(f"new string config value must be None or of type str,"
+                         f" got {new_val} of type {type(new_val)}.")
+
+    return _StateContextManager(name, help, update_thread_local_hook, validate)
 
   def _trace_context(self):
     """Returns a tuple of configuration values that affect tracing.
@@ -295,33 +330,40 @@ class Config:
     return (self.x64_enabled, self.jax_numpy_rank_promotion,
             self.jax_default_matmul_precision)
 
-class _BoolStateContextManager:
-  def __init__(self, name, help, update_thread_local_hook):
+class _StateContextManager:
+  def __init__(self, name, help, update_thread_local_hook,
+               validate_new_val_hook: Optional[Callable[[Any], None]] = None):
     self._name = name
     self.__name__ = name[4:] if name.startswith('jax_') else name
     self.__doc__ = f"Context manager for `{name}` config option.\n\n{help}"
-    self._hook = update_thread_local_hook
+    self._update_thread_local_hook = update_thread_local_hook
+    self._validate_new_val_hook = validate_new_val_hook
 
   @contextlib.contextmanager
-  def __call__(self, new_val: bool):
+  def __call__(self, new_val):
+    if self._validate_new_val_hook:
+      self._validate_new_val_hook(new_val)
     prev_val = getattr(_thread_local_state, self._name, unset)
     setattr(_thread_local_state, self._name, new_val)
-    if self._hook is not None: self._hook(new_val)
+    if self._update_thread_local_hook:
+      self._update_thread_local_hook(new_val)
     try:
       yield
     finally:
       if prev_val is unset:
         delattr(_thread_local_state, self._name)
-        if self._hook is not None: self._hook(None)
+        if self._update_thread_local_hook:
+          self._update_thread_local_hook(None)
       else:
         setattr(_thread_local_state, self._name, prev_val)
-        if self._hook is not None: self._hook(prev_val)
+        if self._update_thread_local_hook:
+          self._update_thread_local_hook(prev_val)
 
   def _add_hooks(self, update_global_hook, update_thread_local_hook):
     """Private method that adds hooks to an existing context-manager.
 
     Used to avoid cyclic import dependencies."""
-    self._hook = update_thread_local_hook
+    self._update_thread_local_hook = update_thread_local_hook
     config._update_hooks[self._name] = update_global_hook
     update_global_hook(config._read(self._name))
 
@@ -407,7 +449,15 @@ flags.DEFINE_integer(
           'until the Python callback consume more outfeeds.'),
     lower_bound=int(16 * 1e6)
 )
-
+flags.DEFINE_bool(
+    'jax_host_callback_outfeed',
+    bool_env('JAX_HOST_CALLBACK_OUTFEED', False),
+    help=(
+        'Use outfeed implementation for host_callback, even on CPU and GPU. '
+        'If false, use the CustomCall implementation. '
+        'Has no effect on TPU, since only the outfeed mechanism is implemented.'
+    )
+)
 
 enable_checks = config.define_bool_state(
     name='jax_enable_checks',
@@ -447,12 +497,22 @@ log_compiles = config.define_bool_state(
           'DEBUG.'))
 
 distributed_debug = config.define_bool_state(
-    name="jax_distributed_debug",
+    name='jax_distributed_debug',
     default=False,
     help=('Enable logging useful for debugging multi-process distributed '
           'computations. Logging is performed with `absl.logging` at WARNING '
           'level.'))
 
+hlo_source_file_canonicalization_regex = config.define_string_state(
+    name='jax_hlo_source_file_canonicalization_regex',
+    default=None,
+    help=('Used to canonicalize the source_path metadata of HLO instructions '
+          'by removing the given regex. If set, re.sub() is called on each '
+          'source_file with the given regex, and all matches are removed. '
+          'This can be used to avoid spurious cache misses when using the '
+          'persistent compilation cache, which includes HLO metadata in the '
+          'cache key.'))
+
 def _update_x64_global(val):
   lib.jax_jit.global_state().enable_x64 = val
 
