@@ -954,13 +954,7 @@ batching.collective_rules[all_to_all_p] = _all_to_all_batched_collective
 core.axis_substitution_rules[all_to_all_p] = partial(_subst_all_names_in_param, 'axis_name')
 
 
-def _expand(dim, size, index, x):
-  shape = list(x.shape)
-  shape.insert(dim, size)
-  out = lax.full(shape, lax._const(x, 0))
-  return lax.dynamic_update_index_in_dim(out, x, index, dim)
-
-def all_gather(x, axis_name, *, axis_index_groups=None):
+def all_gather(x, axis_name, *, axis_index_groups=None, axis=0, tiled=False):
   """Gather values of x across all replicas.
 
   If ``x`` is a pytree then the result is equivalent to mapping this function to
@@ -976,11 +970,21 @@ def all_gather(x, axis_name, *, axis_index_groups=None):
       an axis of size 4, [[0, 1], [2, 3]] would run all gather over the first
       two and last two replicas). Groups must cover all axis indices exactly
       once, and all groups must be the same size.
+    axis: a positional axis into which the chunks along ``axis_name`` will be
+      concatenated.
+    tiled: when ``False``, the chunks will be stacked into a fresh positional
+      axis at index ``axis`` in the output. When ``True``, ``axis`` has to
+      refer to an existing positional dimension and the chunks will be
+      concatenated into that dimension.
 
   Returns:
     Array(s) representing the result of an all-gather along the axis
-    ``axis_name``. Shapes are the same as ``x.shape``, but with a leading
-    dimension of the axis_size.
+    ``axis_name``. Shapes are the same as ``x.shape``, but:
+
+    - when ``tiled`` is ``False``, there is a new dimension equal to the
+      size of axis ``axis_name`` in position ``axis``,
+    - when ``tiled`` is ``True``, the size of dimension in position ``axis``
+      is multiplied by the size of axis ``axis_name``.
 
   For example, with 4 XLA devices available:
 
@@ -1015,45 +1019,64 @@ def all_gather(x, axis_name, *, axis_index_groups=None):
     [ 4  5  6  7]]]
   """
   axis_size = psum(1, axis_name, axis_index_groups=axis_index_groups)
-  bind = partial(all_gather_p.bind, all_gather_dimension=0,
+  bind = partial(all_gather_p.bind, all_gather_dimension=axis,
                  axis_name=axis_name, axis_index_groups=axis_index_groups,
-                 axis_size=axis_size)
+                 axis_size=axis_size, tiled=tiled)
   return tree_util.tree_map(bind, x)
 
-def _all_gather_via_psum(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+def _expand(dim, size, index, tiled, x):
+  shape = list(x.shape)
+  if tiled:
+    tile_size = shape[dim]
+    shape[dim] *= size
+    out = lax.full(shape, lax._const(x, 0))
+    return lax.dynamic_update_slice_in_dim(out, x, index * tile_size, dim)
+  else:
+    shape.insert(dim, size)
+    out = lax.full(shape, lax._const(x, 0))
+    return lax.dynamic_update_index_in_dim(out, x, index, dim)
+
+def _all_gather_via_psum(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled):
   index = _index_in_group(axis_name, axis_index_groups)
-  outs = tree_util.tree_map(partial(_expand, all_gather_dimension, axis_size, index), x)
+  outs = tree_util.tree_map(partial(_expand, all_gather_dimension, axis_size, index, tiled), x)
   return psum(outs, axis_name, axis_index_groups=axis_index_groups)
 
-def _all_gather_impl(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+def _all_gather_impl(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled):
   raise AssertionError("Unexpected call to _all_gather_impl")
 
-def _all_gather_translation_rule(c, x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, axis_env, platform):
+def _all_gather_translation_rule(c, x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled, axis_env, platform):
   # TODO(cjfj): Enable this for TPU also?
   if (platform == 'gpu') and (all_gather_dimension == 0):
-    new_shape = list(c.get_shape(x).dimensions())
-    new_shape.insert(all_gather_dimension, 1)
-    broadcast_dimensions = [i for i in range(len(new_shape)) if i != all_gather_dimension]
-    x = xops.BroadcastInDim(x, new_shape, broadcast_dimensions)
+    if not tiled:
+      new_shape = list(c.get_shape(x).dimensions())
+      new_shape.insert(all_gather_dimension, 1)
+      broadcast_dimensions = [i for i in range(len(new_shape)) if i != all_gather_dimension]
+      x = xops.BroadcastInDim(x, new_shape, broadcast_dimensions)
     replica_groups = _replica_groups(axis_env, axis_name, axis_index_groups)
     return xops.AllGather(x, all_gather_dimension=all_gather_dimension, shard_count=axis_size,
                           replica_groups=xc.make_replica_groups(replica_groups))
   else:
     lowering = xla.lower_fun(_all_gather_via_psum, multiple_results=False, parallel=True)
     return lowering(c, x, all_gather_dimension=all_gather_dimension, axis_name=axis_name,
-                    axis_index_groups=axis_index_groups, axis_size=axis_size, axis_env=axis_env, platform=platform)
+                    axis_index_groups=axis_index_groups, axis_size=axis_size, tiled=tiled,
+                    axis_env=axis_env, platform=platform)
 
-def _all_gather_abstract_eval(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+def _all_gather_abstract_eval(x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled):
   if not isinstance(axis_name, (list, tuple)):
     axis_name = (axis_name,)
   x_aval = raise_to_shaped(x)
   new_shape = list(x_aval.shape)
-  new_shape.insert(all_gather_dimension, axis_size)
+  if tiled:
+    new_shape[all_gather_dimension] *= axis_size
+  else:
+    new_shape.insert(all_gather_dimension, axis_size)
   new_named_shape = {name: size for name, size in x_aval.named_shape.items()
                      if name not in axis_name}
   return x_aval.update(shape=new_shape, named_shape=new_named_shape)
 
-def _all_gather_transpose_rule(cts, x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+def _all_gather_transpose_rule(cts, x, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled):
+  if tiled:
+    raise NotImplementedError("Please open a feature request!")
   # TODO(cjfj): Add reduce-scatter op to XLA?
   concat_axis = 0
   return (lax_numpy.sum(all_to_all(
@@ -1061,7 +1084,9 @@ def _all_gather_transpose_rule(cts, x, *, all_gather_dimension, axis_name, axis_
       concat_axis=concat_axis, axis_index_groups=axis_index_groups),
       axis=concat_axis),)
 
-def _all_gather_batcher(vals_in, dims_in, *, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+def _all_gather_batcher(vals_in, dims_in, *, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled):
+  if tiled:
+    raise NotImplementedError("Please open a feature request!")
   (x,), (d,) = vals_in, dims_in
   if d <= all_gather_dimension:
     all_gather_dimension += 1
@@ -1072,10 +1097,13 @@ def _all_gather_batcher(vals_in, dims_in, *, all_gather_dimension, axis_name, ax
       all_gather_dimension=all_gather_dimension,
       axis_name=axis_name,
       axis_index_groups=axis_index_groups,
-      axis_size=axis_size)
+      axis_size=axis_size,
+      tiled=tiled)
   return result, d
 
-def _all_gather_batched_collective(frame, vals_in, dims_in, all_gather_dimension, axis_name, axis_index_groups, axis_size):
+def _all_gather_batched_collective(frame, vals_in, dims_in, all_gather_dimension, axis_name, axis_index_groups, axis_size, tiled):
+  if tiled:
+    raise NotImplementedError("Please open a feature request!")
   assert axis_index_groups is None, "axis_index_groups not supported in vmap"
   assert axis_size == frame.size, "axis size doesn't match"
   if not isinstance(axis_name, tuple):
@@ -1101,6 +1129,168 @@ batching.primitive_batchers[all_gather_p] = _all_gather_batcher
 batching.collective_rules[all_gather_p] = _all_gather_batched_collective
 core.axis_substitution_rules[all_gather_p] = partial(_subst_all_names_in_param, 'axis_name')
 
+
+def _reduce_scatter_via_reducer(x, *, reducer, scatter_dimension, axis_name, axis_index_groups, axis_size, tiled):
+  index = _index_in_group(axis_name, axis_index_groups)
+  scatter_dim_input_size = x.shape[scatter_dimension]
+  if tiled and scatter_dim_input_size % axis_size != 0:
+      raise ValueError(f"tiled reduce_scatter operand scatter dimension size "
+                       f"{scatter_dim_input_size} must be divisible by "
+                       f"shard count {axis_size}")
+  elif not tiled and scatter_dim_input_size != axis_size:
+      raise ValueError(f"reduce_scatter operand scatter dimension size "
+                       f"{scatter_dim_input_size} must match shard count"
+                       f"{axis_size}")
+  scatter_dim_output_size = scatter_dim_input_size // axis_size
+
+  outs = reducer(x, axis_name=axis_name, axis_index_groups=axis_index_groups)
+  outs = lax.dynamic_slice_in_dim(
+      outs,
+      start_index=index * scatter_dim_output_size,
+      slice_size=scatter_dim_output_size,
+      axis=scatter_dimension)
+  if not tiled:
+    outs = lax.squeeze(outs, [scatter_dimension])
+  return outs
+
+
+def _reduce_scatter_translation_rule(prim, reducer, c, x, *, scatter_dimension, axis_name,axis_index_groups, axis_size, tiled, axis_env, platform):
+  # TODO(b/194706412): Enable this for TPU?
+  if platform == "gpu":
+    scalar = ShapedArray((), c.get_shape(x).numpy_dtype())
+    computation = xla.primitive_subcomputation(prim, scalar, scalar)
+    replica_groups = _replica_groups(axis_env, axis_name, axis_index_groups)
+    x = xops.ReduceScatter(
+        x,
+        computation,
+        scatter_dimension=scatter_dimension,
+        shard_count=axis_size,
+        replica_groups=xc.make_replica_groups(replica_groups))
+    if not tiled:
+      new_shape = list(c.get_shape(x).dimensions())
+      del new_shape[scatter_dimension]
+      x = xops.Reshape(x, new_shape)
+    return x
+  else:
+    return xla.lower_fun(
+        _reduce_scatter_via_reducer, multiple_results=False, parallel=True)(
+            c,
+            x,
+            reducer=reducer,
+            scatter_dimension=scatter_dimension,
+            axis_name=axis_name,
+            axis_index_groups=axis_index_groups,
+            axis_size=axis_size,
+            tiled=tiled,
+            axis_env=axis_env,
+            platform=platform)
+
+
+def _reduce_scatter_abstract_eval(x, *, axis_name, scatter_dimension,
+                                  axis_index_groups, axis_size, tiled):
+  if not isinstance(axis_name, (list, tuple)):
+    axis_name = (axis_name,)
+  x_aval = core.raise_to_shaped(x)
+  new_shape = list(x_aval.shape)
+  scatter_dim_input_size = x_aval.shape[scatter_dimension]
+  if tiled:
+    if scatter_dim_input_size % axis_size != 0:
+      raise ValueError(f"tiled reduce_scatter operand scatter dimension size "
+                       f"{scatter_dim_input_size} must be divisible by "
+                       f"shard_count {axis_size}")
+    new_shape[scatter_dimension] = scatter_dim_input_size // axis_size
+  else:
+    if scatter_dim_input_size != axis_size:
+      raise ValueError(f"reduce_scatter operand scatter dimension size "
+                       f"{scatter_dim_input_size} must match shard count "
+                       f"{axis_size}")
+    del new_shape[scatter_dimension]
+
+  new_named_shape = {
+      name: size
+      for name, size in x_aval.named_shape.items()
+      if name not in axis_name
+  }
+  return x_aval.update(shape=new_shape, named_shape=new_named_shape)
+
+
+reduce_scatter_p = core.AxisPrimitive("reduce_scatter")
+reduce_scatter_p.def_abstract_eval(_reduce_scatter_abstract_eval)
+xla.parallel_translations[reduce_scatter_p] = partial(
+    _reduce_scatter_translation_rule, lax.add_p, psum)
+pxla.multi_host_supported_collectives.add(reduce_scatter_p)
+
+
+def psum_scatter(x, axis_name, *, scatter_dimension=0, axis_index_groups=None, tiled=False):
+  """Compute an all-reduce sum over the axis ``axis_name``, and scatter the result.
+
+  Args:
+    x: array(s) with a mapped axis named ``axis_name``.
+    axis_name: hashable Python object used to name a pmapped axis (see the
+      :func:`jax.pmap` documentation for more details).
+    scatter_dimension: a positional axis into which the all reduce result along
+      ``axis_name`` will be scattered.
+    axis_index_groups: optional list of lists containing axis indices (e.g. for
+      an axis of size 4, [[0, 1], [2, 3]] would run reduce-scatter over the
+      first two and the last two replicas). Groups must cover all axis indices
+      exactly once, and all groups must be the same size.
+    tiled: when ``False``, the size of dimension in ``scatter_dimension`` must
+      match the size of axis ``axis_name`` (or the group size if
+      ``axis_index_groups`` is given). After scattering the all reduce result
+      along ``scatter_dimension``, the output is sequeezed by removing
+      ``scatter_dimension``. When ``True``, the size of dimension in
+      ``scatter_dimension` must be dividible by the size of axis ``axis_name``
+      (or the group size if ``axis_index_groups`` is given),
+      and ``scatter_dimension`` is preserved.
+
+  Returns:
+    Array(s) with the similar shape as ``x``, except the size of dimension in
+    position``scatter_dimension`` is divided by the size of axis ``axis_name``.
+
+  For example, with 4 XLA devices available:
+
+  >>> x = np.arange(16).reshape(4,4)
+  >>> print(x)
+  [[ 0  1  2  3]
+   [ 4  5  6  7]
+   [ 8  9 10 11]
+   [12 13 14 15]]
+  >>> y = jax.pmap(lambda x: jax.lax.psum_scatter(x, 'i'), axis_name='i')(x)
+  >>> print(y)
+  [24 28 32 36]
+
+  if using tiled:
+
+  >>> y = jax.pmap(lambda x: jax.lax.psum_scatter(x, 'i', tiled=True), axis_name='i')(x)
+  >>> print(y)
+  [[24]
+   [28]
+   [32]
+   [36]]
+
+  An example of using axis_index_groups:
+
+  >>> def f(x):
+  ...   return jax.lax.psum_scatter(
+  ...       x, 'i', axis_index_groups=[[0, 2], [3, 1]], tiled=True)
+  >>> y = jax.pmap(f, axis_name='i')(x)
+  >>> print(y)
+  [[ 8 10]
+   [20 22]
+   [12 14]
+   [16 18]]
+  """
+  axis_size = psum(1, axis_name, axis_index_groups=axis_index_groups)
+  bind = partial(
+      reduce_scatter_p.bind,
+      axis_name=axis_name,
+      scatter_dimension=scatter_dimension,
+      axis_index_groups=axis_index_groups,
+      axis_size=axis_size,
+      tiled=tiled)
+  return tree_util.tree_map(bind, x)
+
+
 def _axis_index_translation_rule(c, *, axis_name, axis_env, platform):
   axis_pos = list(axis_env.names).index(axis_name)
   nreplicas = axis_env.nreps // prod(axis_env.sizes)
