@@ -66,8 +66,6 @@ def convolve(in1, in2, mode='full', method='auto',
              precision=None):
   if method != 'auto':
     warnings.warn("convolve() ignores method argument")
-  if jnp.issubdtype(in1.dtype, jnp.complexfloating) or jnp.issubdtype(in2.dtype, jnp.complexfloating):
-    raise NotImplementedError("convolve() does not support complex inputs")
   return _convolve_nd(in1, in2, mode, precision=precision)
 
 
@@ -76,8 +74,6 @@ def convolve2d(in1, in2, mode='full', boundary='fill', fillvalue=0,
                precision=None):
   if boundary != 'fill' or fillvalue != 0:
     raise NotImplementedError("convolve2d() only supports boundary='fill', fillvalue=0")
-  if jnp.issubdtype(in1.dtype, jnp.complexfloating) or jnp.issubdtype(in2.dtype, jnp.complexfloating):
-    raise NotImplementedError("convolve2d() does not support complex inputs")
   if jnp.ndim(in1) != 2 or jnp.ndim(in2) != 2:
     raise ValueError("convolve2d() only supports 2-dimensional inputs.")
   return _convolve_nd(in1, in2, mode, precision=precision)
@@ -88,9 +84,7 @@ def correlate(in1, in2, mode='full', method='auto',
               precision=None):
   if method != 'auto':
     warnings.warn("correlate() ignores method argument")
-  if jnp.issubdtype(in1.dtype, jnp.complexfloating) or jnp.issubdtype(in2.dtype, jnp.complexfloating):
-    raise NotImplementedError("correlate() does not support complex inputs")
-  return _convolve_nd(in1, jnp.flip(in2), mode, precision=precision)
+  return _convolve_nd(in1, jnp.flip(in2.conj()), mode, precision=precision)
 
 
 @_wraps(osp_signal.correlate2d)
@@ -98,11 +92,30 @@ def correlate2d(in1, in2, mode='full', boundary='fill', fillvalue=0,
                 precision=None):
   if boundary != 'fill' or fillvalue != 0:
     raise NotImplementedError("correlate2d() only supports boundary='fill', fillvalue=0")
-  if jnp.issubdtype(in1.dtype, jnp.complexfloating) or jnp.issubdtype(in2.dtype, jnp.complexfloating):
-    raise NotImplementedError("correlate2d() does not support complex inputs")
   if jnp.ndim(in1) != 2 or jnp.ndim(in2) != 2:
-    raise ValueError("correlate2d() only supports {ndim}-dimensional inputs.")
-  return _convolve_nd(in1[::-1, ::-1], in2, mode, precision=precision)[::-1, ::-1]
+    raise ValueError("correlate2d() only supports 2-dimensional inputs.")
+
+  swap = all(s1 <= s2 for s1, s2 in zip(in1.shape, in2.shape))
+  same_shape =  all(s1 == s2 for s1, s2 in zip(in1.shape, in2.shape))
+
+  if mode == "same":
+    in1, in2 = in1[::-1, ::-1], in2.conj()
+    result = _convolve_nd(in1, in2, mode, precision=precision)[::-1, ::-1]
+  elif mode == "valid":
+    if swap and not same_shape:
+      in1, in2 = in2[::-1, ::-1], in1.conj()
+      result = _convolve_nd(in1, in2, mode, precision=precision)
+    else:
+      in1, in2 = in1[::-1, ::-1], in2.conj()
+      result = _convolve_nd(in1, in2, mode, precision=precision)[::-1, ::-1]
+  else:
+    if swap:
+      in1, in2 = in2[::-1, ::-1], in1.conj()
+      result = _convolve_nd(in1, in2, mode, precision=precision).conj()
+    else:
+      in1, in2 = in1[::-1, ::-1], in2.conj()
+      result = _convolve_nd(in1, in2, mode, precision=precision)[::-1, ::-1]
+  return result
 
 
 @_wraps(osp_signal.detrend)
