@@ -27,7 +27,7 @@ from jax._src.numpy.lax_numpy import (asarray, _reduction_dims, _constant_like,
                                       _promote_args_inexact)
 from jax._src.numpy.util import _wraps
 
-from typing import Tuple
+from typing import Optional, Tuple
 
 
 @_wraps(osp_special.gammaln)
@@ -121,7 +121,7 @@ def logsumexp(a, axis=None, b=None, keepdims=False, return_sign=False):
                                   axis=dims, keepdims=keepdims)),
                   amax)
     sign = jnp.where(jnp.isnan(out), np.nan, 1.0).astype(out.dtype)
-    sign = jnp.where(out == -np.inf, 0.0, sign)
+    sign = jnp.where(jnp.isneginf(out), 0.0, sign)
   else:
     expsub = lax.exp(lax.sub(a, amax_with_dims))
     if b is not None:
@@ -177,10 +177,10 @@ def multigammaln(a, d):
   constant = lax.mul(lax.mul(lax.mul(_constant_like(a, 0.25), d_),
                              lax.sub(d_, _constant_like(a, 1))),
                      lax.log(_constant_like(a, np.pi)))
+  b = lax.div(jnp.arange(d, dtype=d_.dtype), _constant_like(a, 2))
   res = jnp.sum(gammaln(jnp.expand_dims(a, axis=-1) -
-                        lax.div(jnp.arange(d, dtype=d_.dtype),
-                                _constant_like(a, 2))),
-               axis=-1)
+                        jnp.expand_dims(b, axis=tuple(range(a.ndim)))),
+                axis=-1)
   return res + constant
 
 
@@ -220,14 +220,16 @@ def zeta(x, q=None):
   # precision ~ N, M
   N = M = dtype(8) if lax.dtype(a) == jnp.float32 else dtype(16)
   assert M <= len(_BERNOULLI_COEFS)
-  k = np.arange(N, dtype=N.dtype)
+  k = jnp.expand_dims(np.arange(N, dtype=N.dtype), tuple(range(a.ndim)))
   S = jnp.sum((a_ + k) ** -s_, -1)
   I = lax.div((a + N) ** (dtype(1) - s), s - dtype(1))
   T0 = (a + N) ** -s
-  s_over_a = (s_ + np.arange(2 * M, dtype=M.dtype)) / (a_ + N)
+  m = jnp.expand_dims(np.arange(2 * M, dtype=M.dtype), tuple(range(s.ndim)))
+  s_over_a = (s_ + m) / (a_ + N)
   T1 = jnp.cumprod(s_over_a, -1)[..., ::2]
   T1 = jnp.clip(T1, a_max=jnp.finfo(dtype).max)
-  coefs = np.array(_BERNOULLI_COEFS[:T1.shape[-1]], dtype=dtype)
+  coefs = np.expand_dims(np.array(_BERNOULLI_COEFS[:T1.shape[-1]], dtype=dtype),
+                         tuple(range(a.ndim)))
   T1 = T1 / coefs
   T = T0 * (dtype(0.5) + T1.sum(-1))
   return S + I + T
@@ -909,7 +911,8 @@ def _gen_associated_legendre(l_max: int,
     p_val = p_val + h
     return p_val
 
-  p = lax.fori_loop(lower=2, upper=l_max+1, body_fun=body_fun, init_val=p)
+  if l_max > 1:
+    p = lax.fori_loop(lower=2, upper=l_max+1, body_fun=body_fun, init_val=p)
 
   return p
 
@@ -1010,3 +1013,77 @@ def lpmn_values(m: int, n: int, z: jnp.ndarray, is_normalized: bool) -> jnp.ndar
   l_max = n
 
   return _gen_associated_legendre(l_max, z, is_normalized)
+
+
+
+@partial(jit, static_argnums=(4,))
+def _sph_harm(m: jnp.ndarray,
+              n: jnp.ndarray,
+              theta: jnp.ndarray,
+              phi: jnp.ndarray,
+              n_max: int) -> jnp.ndarray:
+  """Computes the spherical harmonics."""
+
+  cos_colatitude = jnp.cos(phi)
+
+  legendre = _gen_associated_legendre(n_max, cos_colatitude, True)
+  legendre_val = legendre[abs(m), n, jnp.arange(len(n))]
+
+  angle = abs(m) * theta
+  vandermonde = lax.complex(jnp.cos(angle), jnp.sin(angle))
+  harmonics = lax.complex(legendre_val * jnp.real(vandermonde),
+                          legendre_val * jnp.imag(vandermonde))
+
+  # Negative order.
+  harmonics = jnp.where(m < 0,
+                        (-1.0)**abs(m) * jnp.conjugate(harmonics),
+                        harmonics)
+
+  return harmonics
+
+
+def sph_harm(m: jnp.ndarray,
+             n: jnp.ndarray,
+             theta: jnp.ndarray,
+             phi: jnp.ndarray,
+             n_max: Optional[int] = None) -> jnp.ndarray:
+  r"""Computes the spherical harmonics.
+
+  The JAX version has one extra argument `n_max`, the maximum value in `n`.
+
+  The spherical harmonic of degree `n` and order `m` can be written as
+  :math:`Y_n^m(\theta, \phi) = N_n^m * P_n^m(\cos \phi) * \exp(i m \theta)`,
+  where :math:`N_n^m = \sqrt{\frac{\left(2n+1\right) \left(n-m\right)!}
+  {4 \pi \left(n+m\right)!}}` is the normalization factor and :math:`\phi` and
+  :math:\theta` are the colatitude and longitude, repectively. :math:`N_n^m` is
+  chosen in the way that the spherical harmonics form a set of orthonormal basis
+  functions of :math:`L^2(S^2)`.
+
+  Args:
+    m: The order of the harmonic; must have `|m| <= n`. Return values for
+      `|m| > n` ara undefined.
+    n: The degree of the harmonic; must have `n >= 0`. The standard notation for
+      degree in descriptions of spherical harmonics is `l (lower case L)`. We
+      use `n` here to be consistent with `scipy.special.sph_harm`. Return
+      values for `n < 0` are undefined.
+    theta: The azimuthal (longitudinal) coordinate; must be in [0, 2*pi].
+    phi: The polar (colatitudinal) coordinate; must be in [0, pi].
+    n_max: The maximum degree `max(n)`. If the supplied `n_max` is not the true
+      maximum value of `n`, the results are clipped to `n_max`. For example,
+      `sph_harm(m=jnp.array([2]), n=jnp.array([10]), theta, phi, n_max=6)`
+      acutually returns
+      `sph_harm(m=jnp.array([2]), n=jnp.array([6]), theta, phi, n_max=6)`
+  Returns:
+    A 1D array containing the spherical harmonics at (m, n, theta, phi).
+  """
+
+  if jnp.isscalar(phi):
+    phi = jnp.array([phi])
+
+  if n_max is None:
+    n_max = jnp.max(n)
+  n_max = core.concrete_or_error(
+      int, n_max, 'The `n_max` argument of `jnp.scipy.special.sph_harm` must '
+      'be statically specified to use `sph_harm` within JAX transformations.')
+
+  return _sph_harm(m, n, theta, phi, n_max)
