@@ -18,6 +18,7 @@ from operator import attrgetter
 from contextlib import contextmanager
 from collections import namedtuple
 from functools import total_ordering
+import gc
 import itertools as it
 from weakref import ref
 import threading
@@ -32,7 +33,7 @@ from ._src import dtypes
 from ._src import config as jax_config
 from ._src.config import FLAGS, config
 from .errors import (ConcretizationTypeError, TracerArrayConversionError,
-                     TracerIntegerConversionError)
+                     TracerIntegerConversionError, UnexpectedTracerError)
 from . import linear_util as lu
 
 from jax._src import source_info_util
@@ -306,11 +307,33 @@ def extract_call_jaxpr(
     return (params["call_jaxpr"], new_params)
 
 
+# TODO(mattjj): replace this approach with a primitive-keyed table of rules
 def traverse_jaxpr_params(f, params):
   """Applies f to each jaxpr parameter and returns a tuple of returned values."""
-  return tuple(f(param if type(param) is Jaxpr else param.jaxpr)
-               for param in params.values()
-               if type(param) in (Jaxpr, ClosedJaxpr))
+  return {name: f(p)
+          for name, param in params.items()
+          for p in (param if isinstance(param, (tuple, list)) else [param])
+          if type(p) in (Jaxpr, ClosedJaxpr)}
+
+
+def eval_jaxpr_eqn(eqn, in_vals):
+  """Evaluates the jaxpr equation with the provided input values."""
+  call_jaxpr, params = extract_call_jaxpr(eqn.primitive, eqn.params)
+  if call_jaxpr:
+    subfuns = [lu.wrap_init(partial(eval_jaxpr, call_jaxpr, ()))]
+  else:
+    subfuns = []
+  if eqn.primitive in initial_to_final_param_rules:
+    bind_params = initial_to_final_param_rules[eqn.primitive](params)
+  elif eqn.primitive.map_primitive:
+    out_axes_thunk = HashableFunction(lambda: params['out_axes'],
+                                      closure=params['out_axes'])
+    bind_params = dict(params, out_axes_thunk=out_axes_thunk)
+    del bind_params['out_axes']
+  else:
+    bind_params = params
+  with source_info_util.user_context(eqn.source_info):
+    return eqn.primitive.bind(*(subfuns + in_vals), **bind_params)
 
 
 def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
@@ -328,23 +351,7 @@ def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
   map(write, jaxpr.constvars, consts)
   map(write, jaxpr.invars, args)
   for eqn in jaxpr.eqns:
-    in_vals = map(read, eqn.invars)
-    call_jaxpr, params = extract_call_jaxpr(eqn.primitive, eqn.params)
-    if call_jaxpr:
-      subfuns = [lu.wrap_init(partial(eval_jaxpr, call_jaxpr, ()))]
-    else:
-      subfuns = []
-    if eqn.primitive in initial_to_final_param_rules:
-      bind_params = initial_to_final_param_rules[eqn.primitive](params)
-    elif eqn.primitive.map_primitive:
-      out_axes_thunk = HashableFunction(lambda: params['out_axes'],
-                                        closure=params['out_axes'])
-      bind_params = dict(params, out_axes_thunk=out_axes_thunk)
-      del bind_params['out_axes']
-    else:
-      bind_params = params
-    with source_info_util.user_context(eqn.source_info):
-      ans = eqn.primitive.bind(*(subfuns + in_vals), **bind_params)
+    ans = eval_jaxpr_eqn(eqn, map(read, eqn.invars))
     if eqn.primitive.multiple_results:
       map(write, eqn.outvars, ans)
     else:
@@ -433,34 +440,34 @@ class Trace:
 
 def escaped_tracer_error(tracer, detail=None):
   num_frames = FLAGS.jax_tracer_error_num_traceback_frames
-  msg = ("Encountered an unexpected tracer. Perhaps this tracer escaped "
-         "through global state from a previously traced function.\n"
-         "The functions being transformed should not save traced values to "
-         "global state.")
-  if detail:
-    msg += " Detail: {}.".format(detail)
-  try:
-    line_info = tracer._line_info
-  except AttributeError:
-    pass
-  else:
-    msg += ('\nThe tracer that caused this error was created on line '
-            f'{source_info_util.summarize(line_info)}.\n')
-    if num_frames > 0:
-      msg += (f'When the tracer was created, the final {num_frames} stack '
-              'frames (most recent last) excluding JAX-internal frames were:\n'
-              f'{source_info_util.summarize(line_info, num_frames=num_frames)}')
+  msg = ('Encountered an unexpected tracer. A function transformed by JAX '
+         'had a side effect, allowing for a reference to an intermediate value '
+         f'with shape {tracer.shape} and dtype {tracer.dtype} to escape.\n'
+         'JAX transformations require that functions explicitly return their '
+         'outputs, and disallow saving intermediate values to global state.')
   dbg = getattr(tracer._trace.main, 'debug_info', None)
   if dbg is not None:
-    msg += ('\nThe function being traced when the tracer leaked was '
+    msg += ('\nThe function being traced when the value leaked was '
             f'{dbg.func_src_info} traced for {dbg.traced_for}.')
+  line_info = getattr(tracer, '_line_info', None)
+  if line_info is not None:
+    divider = '\n' + '-'*30 + '\n'
+    msg += divider
+    msg += ('The leaked intermediate value was created on line '
+            f'{source_info_util.summarize(line_info)}. ')
+    msg += divider
+    if num_frames > 0:
+      msg += (f'When the value was created, the final {num_frames} stack '
+              'frames (most recent last) excluding JAX-internal frames were:')
+      msg += divider + source_info_util.summarize(
+          line_info, num_frames=num_frames) + divider
   msg += ('\nTo catch the leak earlier, try setting the environment variable '
           'JAX_CHECK_TRACER_LEAKS or using the `jax.checking_leaks` context '
           'manager.')
+  if detail:
+    msg += f'Detail: {detail}'
   return UnexpectedTracerError(msg)
 
-class UnexpectedTracerError(Exception): pass
-
 class Tracer:
   __array_priority__ = 1000
   __slots__ = ['_trace', '__weakref__', '_line_info']
@@ -567,7 +574,7 @@ class Tracer:
 
   def __repr__(self):
     base = pp('Traced<{}>with<{}>'.format(self.aval, self._trace))
-    contents = self._contents()
+    contents = [(name, pp(repr(attr))) for name, attr in self._contents()]
     if contents:
       base += pp('  with ') >> vcat(pp('{} = '.format(name)) >> pp_payload
                                     for name, pp_payload in contents)
@@ -575,7 +582,7 @@ class Tracer:
 
   def _contents(self):
     try:
-      return [(name, pp(repr(getattr(self, name)))) for name in self.__slots__]
+      return [(name, getattr(self, name)) for name in self.__slots__]
     except AttributeError:
       return ()
 
@@ -709,14 +716,23 @@ class TraceState:
     return new
 
 
+def _update_thread_local_jit_state(dynamic):
+  # Copies the MainTrace instance, removing any .debug_info or .jaxpr_stack
+  # fields that should not be kept alive as part of a cache key.
+  # TODO(mattjj): split debug_info and jaxpr_stack out of MainTrace.
+  # TODO(mattjj): add a test that verifies that JIT-ted functions are not kept
+  # alive by the JIT cache, particularly for nested JIT-ted functions.
+  copy = MainTrace(dynamic.level, dynamic.trace_type, **dynamic.payload)
+  jax_config.update_thread_local_jit_state(dynamic_trace_state=copy)
+
+
 # The global state of the tracer is accessed by a thread-local object.
 # This allows concurrent tracing in separate threads; passing traced objects
 # between threads is forbidden.
 class ThreadLocalState(threading.local):
   def __init__(self):
     self.trace_state = TraceState()
-    jax_config.update_thread_local_jit_state(
-        dynamic_trace_state=self.trace_state.trace_stack.dynamic)
+    _update_thread_local_jit_state(self.trace_state.trace_stack.dynamic)
 thread_local_state = ThreadLocalState()
 
 def trace_state_clean() -> bool:
@@ -737,6 +753,17 @@ def reset_trace_state() -> bool:
 def cur_sublevel() -> Sublevel:
   return thread_local_state.trace_state.substack[-1]
 
+def maybe_find_leaked_tracers(x: Optional[Union[MainTrace, Sublevel]]):
+  """Find the leaked tracers holding a reference to the MainTrace or SubLevel.
+
+  It's possible there's none! eg. there's some cases where JAX itself holds a
+  reference to `x` inside of a lambda closure, and no tracers were leaked
+  by the user. In this case an empty list is returned.
+  """
+  traces = list(filter(lambda x: isinstance(x, Trace), gc.get_referrers(x)))
+  tracers = list(filter(lambda x: isinstance(x, Tracer), gc.get_referrers(*traces)))
+  return tracers
+
 @contextmanager
 def new_main(trace_type: Type[Trace],
              dynamic: bool = False,
@@ -748,7 +775,7 @@ def new_main(trace_type: Type[Trace],
   stack.push(main)
   if dynamic:
     prev_dynamic, stack.dynamic = stack.dynamic, main
-    jax_config.update_thread_local_jit_state(dynamic_trace_state=stack.dynamic)
+    _update_thread_local_jit_state(stack.dynamic)
 
   try:
     yield main
@@ -756,13 +783,15 @@ def new_main(trace_type: Type[Trace],
     stack.pop()
     if dynamic:
       stack.dynamic = prev_dynamic
-      jax_config.update_thread_local_jit_state(dynamic_trace_state=stack.dynamic)
+      _update_thread_local_jit_state(stack.dynamic)
 
   if config.jax_check_tracer_leaks:
     t = ref(main)
     del main
     if t() is not None:
-      raise Exception(f'Leaked trace {t()}')
+      leaked_tracers = maybe_find_leaked_tracers(t())
+      if leaked_tracers:
+        raise Exception(f'Leaked level {t()}. Leaked tracer(s): {leaked_tracers}.')
 
 @contextmanager
 def new_base_main(trace_type: Type[Trace]) -> Generator[MainTrace, None, None]:
@@ -771,19 +800,21 @@ def new_base_main(trace_type: Type[Trace]) -> Generator[MainTrace, None, None]:
   main = MainTrace(0, trace_type)
   prev_dynamic, stack.dynamic = stack.dynamic, main
   prev_base, stack.stack[0] = stack.stack[0], main
-  jax_config.update_thread_local_jit_state(dynamic_trace_state=stack.dynamic)
+  _update_thread_local_jit_state(stack.dynamic)
   try:
     yield main
   finally:
     stack.dynamic = prev_dynamic
     stack.stack[0] = prev_base
-    jax_config.update_thread_local_jit_state(dynamic_trace_state=stack.dynamic)
+    _update_thread_local_jit_state(stack.dynamic)
 
   if config.jax_check_tracer_leaks:
     t = ref(main)
     del main
     if t() is not None:
-      raise Exception('Leaked trace {}'.format(t()))
+      leaked_tracers = maybe_find_leaked_tracers(t())
+      if leaked_tracers:
+        raise Exception(f'Leaked level {t()}. Leaked tracer(s): {leaked_tracers}.')
 
 @contextmanager
 def eval_context():
@@ -803,7 +834,9 @@ def new_sublevel() -> Generator[None, None, None]:
     t = ref(sublevel)
     del sublevel
     if t() is not None:
-      raise Exception(f'Leaked sublevel {t()}.')
+      leaked_tracers = maybe_find_leaked_tracers(t())
+      if leaked_tracers:
+        raise Exception(f'Leaked sublevel {t()}. Leaked tracer(s): {leaked_tracers}.')
 
 def full_lower(val):
   if isinstance(val, Tracer):
@@ -1039,7 +1072,7 @@ class ShapedArray(UnshapedArray):
   array_abstraction_level = 1
 
   def __init__(self, shape, dtype, weak_type=False, named_shape={}):
-    super(ShapedArray, self).__init__(dtype, weak_type=weak_type)
+    super().__init__(dtype, weak_type=weak_type)
     self.shape = canonicalize_shape(shape)
     self.named_shape = dict(named_shape)
 
@@ -1080,7 +1113,7 @@ class ShapedArray(UnshapedArray):
                        self.weak_type, self.named_shape)
 
   def join(self, other):
-    if self.shape == other.shape and self.dtype == other.dtype:
+    if symbolic_equal_shape(self.shape, other.shape) and self.dtype == other.dtype:
       weak_type = self.weak_type and other.weak_type
       named_shape = join_named_shapes(self.named_shape, other.named_shape)
       return self.update(weak_type=weak_type, named_shape=named_shape)
@@ -1118,8 +1151,8 @@ class ConcreteArray(ShapedArray):
   array_abstraction_level = 0
 
   def __init__(self, val, weak_type=False):
-    super(ConcreteArray, self).__init__(np.shape(val), np.result_type(val),
-                                        weak_type=weak_type)
+    super().__init__(np.shape(val), np.result_type(val),
+                     weak_type=weak_type)
     # Note: canonicalized self.dtype doesn't necessarily match self.val
     self.val = val
     assert self.dtype != np.dtype('O'), val
@@ -1282,6 +1315,9 @@ class DimensionHandler:
     """Implements `0 if d == 0 else 1 + dilation * (d - 1))`"""
     return 0 if d == 0 else 1 + dilation * (d - 1)
 
+  def as_value(self, d: DimSize):
+    """Turns a dimension size into a JAX value that we can compute with."""
+    return d
 
 _dimension_handler_int = DimensionHandler()
 _SPECIAL_DIMENSION_HANDLERS: Dict[type, DimensionHandler] = {}
@@ -1377,6 +1413,11 @@ def stride_shape(s: Shape, window_size: Shape, window_stride: Shape) -> Shape:
   """(s - window_size) // window_stride + 1"""
   return tuple(map(stride_dim, s, window_size, window_stride))
 
+def dimension_as_value(d: DimSize):
+  """Turns a dimension size into a JAX value that we can compute with.
+     This is the identity function for constant dimensions."""
+  handler, ds = _dim_handler_and_canonical(d)
+  return handler.as_value(*ds)
 
 def _canonicalize_dimension(dim: DimSize) -> DimSize:
   if type(dim) in _SPECIAL_DIMENSION_HANDLERS:
@@ -1384,29 +1425,42 @@ def _canonicalize_dimension(dim: DimSize) -> DimSize:
   else:
     return operator.index(dim)
 
-def canonicalize_shape(shape: Shape) -> Shape:
+def canonicalize_shape(shape: Shape, context: str="") -> Shape:
   """Canonicalizes and checks for errors in a user-provided shape value.
 
   Args:
     shape: a Python value that represents a shape.
 
   Returns:
-    A tuple of integers.
+    A tuple of canonical dimension values.
   """
   try:
     return tuple(map(_canonicalize_dimension, shape))
   except TypeError:
     pass
-  raise _invalid_shape_error(shape)
+  raise _invalid_shape_error(shape, context)
 
-def _invalid_shape_error(shape: Shape):
+def canonicalize_dim(d: DimSize, context: str="") -> DimSize:
+  """Canonicalizes and checks for errors in a user-provided shape dimension value.
+
+  Args:
+    f: a Python value that represents a dimension.
+
+  Returns:
+    A canonical dimension value.
+  """
+  return canonicalize_shape((d,), context)[0]
+
+def _invalid_shape_error(shape: Shape, context: str=""):
   msg = ("Shapes must be 1D sequences of concrete values of integer type, "
-         "got {}.")
+         f"got {shape}.")
+  if context:
+    msg += f" {context}."
   if any(isinstance(x, Tracer) and isinstance(get_aval(x), ShapedArray)
          and not isinstance(get_aval(x), ConcreteArray) for x in shape):
     msg += ("\nIf using `jit`, try using `static_argnums` or applying `jit` to "
             "smaller subfunctions.")
-  return TypeError(msg.format(shape))
+  return TypeError(msg)
 
 # ------------------- Named shapes -------------------
 
