@@ -18,10 +18,12 @@ For usage instructions, read the jax2tf.convert docstring, and the
 
 """
 import collections
+import dataclasses
 import itertools
 import functools
 import operator as op
-from typing import Any, Dict, Optional, Sequence, Set, Tuple, Union
+import re
+from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union
 
 
 import jax
@@ -29,11 +31,14 @@ from jax._src.numpy import lax_numpy
 import opt_einsum
 from jax import config
 from jax import core
+from . import jax2tf as jax2tf_internal
+
 import numpy as np
+import tensorflow as tf  # type: ignore[import]
 
 DimSize = core.DimSize
 Shape = core.Shape
-
+TfVal = Any
 
 class InconclusiveDimensionOperation(core.InconclusiveDimensionOperation):
   """Raised when we cannot conclusively compute with symbolic dimensions."""
@@ -58,13 +63,13 @@ class _DimMon(dict):
 
   The representation is a dictionary mapping var:exponent.
   The `var` are strings and the exponents are >= 1.
-  The shape variables are assumed to range over integers >= 1.
+  The dimension variables are assumed to range over integers >= 1.
   """
   def __hash__(self):
     return hash(frozenset(self.items()))
 
   def __str__(self):
-    return ' '.join(f'{key}^{exponent}' if exponent != 1 else str(key)
+    return "*".join(f"{key}^{exponent}" if exponent != 1 else str(key)
                     for key, exponent in sorted(self.items()))
 
   @classmethod
@@ -119,11 +124,25 @@ class _DimMon(dict):
       elif diff > 0: d[key] = diff
     return _DimMon(d)
 
+  def evaluate(self, env: Dict[str, Any]) -> Any:
+    prod = lambda xs: functools.reduce(op.mul, xs) if xs else 1
+    def pow_opt(v, p):
+      return v if p == 1 else pow(v, p)
+    return prod([pow_opt(env[id], deg) for id, deg in self.items()])
+
+
+def _multiply(coeff, v: TfVal) -> TfVal:
+  try:
+    coeff = int(coeff)
+  except:
+    return coeff * v
+  else:
+    return 0 if coeff == 0 else v if coeff == 1 else coeff * v
 
 class _DimPolynomial(dict):
   """Polynomial with integer coefficients for polymorphic shapes.
 
-  The shape variables are assumed to range over integers >= 1.
+  The dimension variables are assumed to range over integers >= 1.
 
   We overload integer operations, but we do that soundly, raising
   :class:`InconclusiveDimensionOperation` when the result is not
@@ -184,8 +203,14 @@ class _DimPolynomial(dict):
     return hash(tuple(sorted(self.items())))
 
   def __str__(self):
-    return ' + '.join(f'{c} {mon}' if c != 1 or mon.degree == 0 else str(mon)
-                      for mon, c in sorted(self.items(), reverse=True)).strip()
+    def _one_monomial(mon, c):
+      if mon.degree == 0:
+        return str(c)
+      if c == 1:
+        return str(mon)
+      return f"{c}*{mon}"
+    return " + ".join(_one_monomial(mon, c)
+                      for mon, c in sorted(self.items(), reverse=True))
 
   def __repr__(self):
     return str(self)
@@ -211,6 +236,14 @@ class _DimPolynomial(dict):
       coeffs[mon] = coeffs.get(mon, 0) + coeff1 * coeff2
     return _DimPolynomial.from_coeffs(coeffs)
 
+  def __pow__(self, power, modulo=None):
+    assert modulo is None
+    try:
+      power = int(power)
+    except:
+      raise InconclusiveDimensionOperation(f"Dimension polynomial cannot be raised to non-integer power '{self}' ^ '{power}'")
+    return functools.reduce(op.mul, [self] * power)
+
   def __rmul__(self, other: DimSize) -> DimSize:
     return self * other  # multiplication commutes
 
@@ -268,7 +301,7 @@ class _DimPolynomial(dict):
     err_msg = f"Dimension polynomial '{self}' is not a multiple of '{divisor}'"
     # invariant: self = dividend + divisor * quotient
     # the leading term of dividend decreases through the loop.
-    while not (isinstance(dividend, int) or dividend.is_constant):
+    while is_poly_dim(dividend) and not dividend.is_constant:
       mon, count = dividend.leading_term
       try:
         qmon = mon.divide(dmon)
@@ -302,6 +335,19 @@ class _DimPolynomial(dict):
   def __rfloordiv__(self, other):
     return _ensure_poly(other).__floordiv__(self)
 
+  def __truediv__(self, divisor: DimSize):
+    # Used for "/"
+    q, r = self.divmod(divisor)
+    if r != 0:
+      raise InconclusiveDimensionOperation(
+          f"Dimension polynomial '{self}' is not a multiple of '{divisor}'")
+    return q
+
+  def __rtruediv__(self, dividend: DimSize):
+    # Used for "/", when dividend is not a _DimPolynomial
+    raise InconclusiveDimensionOperation(
+        f"Division of '{dividend}' by dimension polynomial '{self}' is not supported")
+
   def __mod__(self, divisor: DimSize) -> int:
     return self.divmod(divisor)[1]
 
@@ -330,16 +376,9 @@ class _DimPolynomial(dict):
     return lb, ub
 
   def evaluate(self, env: Dict[str, Any]) -> Any:
-    prod = lambda xs: functools.reduce(op.mul, xs) if xs else 1
-    def mul(coeff, mon):
-      try:
-        coeff = int(coeff)
-      except:
-        return coeff * mon
-      else:
-        return 0 if coeff == 0 else mon if coeff == 1 else coeff * mon
-    terms = [mul(coeff, prod([pow(env[id], deg) for id, deg in mon.items()]))
-             for mon, coeff in self.items()]
+    def pow_opt(v, p):
+      return v if p == 1 else pow(v, p)
+    terms = [_multiply(coeff, mon.evaluate(env)) for mon, coeff in self.items()]
     return sum(terms) if len(terms) > 1 else terms[0]
 
   @property
@@ -403,6 +442,10 @@ class DimensionHandlerPoly(core.DimensionHandler):
           f"window_size '{window_size}', stride '{window_stride}'. Reason: {e}.")
     return d
 
+  def as_value(self, d: DimSize):
+    """Turns a dimension size into a Jax value that we can compute with."""
+    return _dim_as_value(d)
+
 core._SPECIAL_DIMENSION_HANDLERS[_DimPolynomial] = DimensionHandlerPoly()
 
 def _einsum_contract_path(*operands, **kwargs):
@@ -419,7 +462,8 @@ def _einsum_contract_path(*operands, **kwargs):
   # sizes of operands and intermediate results.
   fake_ops = []
   for operand in operands:
-    if isinstance(operand, str):
+    # We replace only array operands
+    if not hasattr(operand, "dtype"):
       fake_ops.append(operand)
     else:
       shape = np.shape(operand)
@@ -438,10 +482,6 @@ def _einsum_contract_path(*operands, **kwargs):
 
   contract_fake_ops, contractions = opt_einsum.contract_path(*fake_ops,
                                                              **kwargs)
-  if len(contractions) > 1:
-    msg = ("Shape polymorphism is not yet supported for einsum with more than "
-           f"one contraction {contractions}")
-    raise ValueError(msg)
   contract_operands = []
   for operand in contract_fake_ops:
     idx = tuple(i for i, fake_op in enumerate(fake_ops) if operand is fake_op)
@@ -451,6 +491,27 @@ def _einsum_contract_path(*operands, **kwargs):
 
 lax_numpy._polymorphic_einsum_contract_path_handlers[_DimPolynomial] = _einsum_contract_path
 
+# A JAX primitive with no array arguments but with a dimension parameter
+# that is a DimPoly. The value of the primitive is the value of the dimension.
+# This primitive is used only in the context of jax2tf, so it does not need
+# XLA translation rules.
+dim_as_value_p = core.Primitive("dim_as_value")
+def _dim_as_value_abstract(dim: DimSize) -> core.AbstractValue:
+  return core.ShapedArray((), np.int32)
+
+dim_as_value_p.def_abstract_eval(_dim_as_value_abstract)
+
+def _dim_as_value(dim: DimSize):
+  return dim_as_value_p.bind(dim=dim)
+
+def _dim_as_value_jax2tf(dim: DimSize):
+  dim_tf, = jax2tf_internal._eval_shape((dim,))
+  assert dim_tf.dtype == tf.int32
+  return dim_tf
+
+def _register_conversion_rules():
+  jax2tf_internal.tf_impl[dim_as_value_p] = _dim_as_value_jax2tf
+
 
 class PolyShape(tuple):
   """Tuple of polymorphic dimension specifications.
@@ -458,6 +519,11 @@ class PolyShape(tuple):
   See docstring of :func:`jax2tf.convert`.
   """
   def __new__(cls, *dim_specs):
+    for i, ds in enumerate(dim_specs):
+      if not isinstance(ds, (int, str)) and ds != ...:
+        msg = (f"Invalid polymorphic shape element: {repr(ds)}; must be a string "
+               "representing a dimension variable, or an integer, or ...")
+        raise ValueError(msg)
     return tuple.__new__(PolyShape, dim_specs)
 
 
@@ -476,82 +542,212 @@ def parse_spec(spec: Optional[Union[str, PolyShape]],
   if spec is None:
     spec_tuple = (...,)  # type: Tuple[Any,...]
   elif isinstance(spec, PolyShape):
-    spec_tuple = tuple(spec)
+    spec_tuple = spec
   elif isinstance(spec, str):
-    spec_ = spec.replace(" ", "")
+    spec_ = spec.strip()
     if spec_[0] == "(":
       if spec_[-1] != ")":
-        raise ValueError(f"PolyShape '{spec}' has invalid syntax")
+        raise ValueError(f"polymorphic shape {repr(spec)} has invalid syntax")
       spec_ = spec_[1:-1]
+      spec_ = spec_.strip()
     spec_ = spec_.rstrip(",")
     if not spec_:
       spec_tuple = ()
     else:
-      specs = spec_.split(',')
-      def parse_dim(ds: str):
-        if ds == "...":
-          return ...
-        elif ds.isdigit():
-          return int(ds)
-        elif ds == "_" or ds.isalnum():
-          return ds
-        else:
-          raise ValueError(f"PolyShape '{spec}' has invalid syntax")
-
-      spec_tuple = tuple(map(parse_dim, specs))
+      spec_tuple = spec_.split(",")  # type: ignore
   else:
-    raise ValueError(f"PolyShape '{spec}' must be either None, a string, or PolyShape.")
+    raise ValueError(f"polymorphic shape {repr(spec)} must be either None, a string, or PolyShape.")
 
+  # Process ...
+  spec_tuple = tuple(map(lambda s: ... if isinstance(s, str) and s.strip() == "..." else s,
+                         spec_tuple))
   ds_ellipses = tuple(ds for ds in spec_tuple if ds == ...)
   if ds_ellipses:
     if len(ds_ellipses) > 1 or spec_tuple[-1] != ...:
-      raise ValueError(f"PolyShape '{spec}' can contain Ellipsis only at the end.")
+      raise ValueError(f"polymorphic shape {repr(spec)} can contain Ellipsis only at the end.")
     spec_tuple = spec_tuple[0:-1]
     if len(arg_shape) >= len(spec_tuple):
       spec_tuple = spec_tuple + ("_",) * (len(arg_shape) - len(spec_tuple))
 
   if len(arg_shape) != len(spec_tuple):
-    raise ValueError(f"PolyShape '{spec}' must match the rank of arguments {arg_shape}.")
+    raise ValueError(f"polymorphic shape {repr(spec)} of rank {len(spec_tuple)} must match the rank {len(arg_shape)} of argument shape {arg_shape}.")
 
-  shape_var_map: Dict[str, Set[int]] = collections.defaultdict(set)
-  def _process_dim(i: int, dim_spec):
-    if not isinstance(dim_spec, (str, int)):
-      raise ValueError(f"PolyShape '{spec}' in axis {i} must contain only integers, strings, or Ellipsis.")
+  # The actual parsing.
+  # We actually parse not just dimension variables, but polynomials.
+  # This is not a supported feature of the API, but is needed when parsing the
+  # polymorphic_shapes of a gradient function, when the primal function has polynomial
+  # output shapes.
+  def _parse_dim(dim_spec: Union[str, int]) -> DimSize:
+    if isinstance(dim_spec, int):
+      return dim_spec  #
+    dim_spec = dim_spec.strip()
+    if not dim_spec:
+      raise ValueError(f"polymorphic shape {repr(spec)} has invalid syntax (empty dimension {dim_spec}')")
+    # Terms are separated by "+"
+    terms = dim_spec.split("+")
+    if not terms:
+      raise ValueError(f"polymorphic shape {repr(spec)} has invalid syntax (empty dimension {dim_spec}')")
+    def _parse_term(term_spec: str) -> DimSize:
+      term_spec = term_spec.strip()
+      # Factors are separated by "*"
+      factors = term_spec.split("*")
+      if not factors:
+        raise ValueError(f"polymorphic shape {repr(spec)} has invalid syntax (unexpected term '{term_spec}')")
+      def _parse_factor(factor_spec: str) -> DimSize:
+        factor_spec = factor_spec.strip()
+        if re.match(r"^-?\d+$", factor_spec):
+          return int(factor_spec)
+        m = re.match(r"^([a-zA-Z]\w*)(\^(\d+))?$", factor_spec)
+        if not m:
+          raise ValueError(f"polymorphic shape {repr(spec)} has invalid syntax (unexpected term '{factor_spec}')")
+        var = _DimPolynomial.from_var(m.group(1))
+        if m.group(3) is None:
+          return var
+        return var ** int(m.group(3))
+
+      return functools.reduce(op.mul, map(_parse_factor, factors))
+    return functools.reduce(op.add, map(_parse_term, terms))
+
+  def _process_dim(i: int, dim_spec: Union[str, int]):
+    if isinstance(dim_spec, str):
+      dim_spec = dim_spec.strip()
     dim_size = arg_shape[i]
+    if isinstance(dim_size, tf.compat.v1.Dimension):
+      dim_size = dim_size.value
     if dim_size is None:
-      if dim_spec == "_" or not isinstance(dim_spec, str):
-        msg = (f"PolyShape '{spec}' in axis {i} must contain a shape variable "
+      def need_dim_var_msg():
+        msg = (f"polymorphic shape {repr(spec)} in axis {i} must contain a dimension variable "
                f"for unknown dimension in argument shape {arg_shape}")
-        raise ValueError(msg)
-      return _DimPolynomial.from_var(dim_spec)
+        if spec is None:
+          msg += ". Perhaps you forgot to add the polymorphic_shapes= parameter to jax2tf.convert?"
+        return msg
+
+      if dim_spec == "_":
+        raise ValueError(need_dim_var_msg())
+      dim_poly = _parse_dim(dim_spec)
+      if not is_poly_dim(dim_poly):
+        raise ValueError(need_dim_var_msg())
+      return dim_poly
     else:  # dim_size is known
+      dim_size = int(dim_size)
       if dim_spec == "_":
         return dim_size
-      if isinstance(dim_spec, int):
-        if dim_spec != dim_size:
-          msg = (f"PolyShape '{spec}' in axis {i} must contain a constant or '_' "
-                 f"for known dimension in argument shape {arg_shape}")
+      dim_poly = _parse_dim(dim_spec)
+      if not is_poly_dim(dim_poly):
+        if dim_poly != dim_size:
+          msg = (f"polymorphic shape {repr(spec)} in axis {i} must match the "
+                 f"known dimension size {dim_size} for argument shape {arg_shape}")
           raise ValueError(msg)
         return dim_size
-      # We have a dimension variable for a known dimension.
-      shape_var_map[dim_spec].add(dim_size)
-      return _DimPolynomial.from_var(dim_spec)
+      return dim_poly
 
   dims = tuple([_process_dim(i, ds) for i, ds in enumerate(spec_tuple)])
-  for dim_var, dim_var_values in shape_var_map.items():
-    if len(dim_var_values) != 1:
-      msg = (f"PolyShape '{spec}' has dimension variable '{dim_var}' "
-             f"corresponding to multiple values ({sorted(dim_var_values)}), for "
-             f"argument shape {arg_shape}")
-      raise ValueError(msg)
-    elif list(dim_var_values)[0] <= 0:
-      msg = (f"PolyShape '{spec}' has dimension variable '{dim_var}' "
-             f"corresponding to 0, for argument shape {arg_shape}")
-      raise ValueError(msg)
-
   return dims
 
+@dataclasses.dataclass
+class DimEquation:
+  # Represents poly == tf_expr
+  poly: _DimPolynomial
+  tf_expr: TfVal
 
-def eval_shape(shape: Sequence[DimSize], shape_env: Dict[str, Any]) -> Sequence[Any]:
+# A dimension environment maps dimension variables to TF expressions that
+# compute the value of the dimension. These expressions refer to the TF
+# function arguments.
+ShapeEnv = Dict[str, TfVal]
+
+def eval_shape(shape: Sequence[DimSize], shape_env: Dict[str, TfVal]) -> Sequence[TfVal]:
   return tuple(d.evaluate(shape_env) if type(d) is _DimPolynomial else d   # type: ignore
                for d in shape)
+
+def solve_dim_equations(eqns: List[DimEquation]) -> ShapeEnv:
+  # Returns a shape environment if it can solve all dimension variables.
+  # Raises an exception if it cannot.
+  shapeenv: ShapeEnv = {}
+
+  def _shapeenv_to_str() -> str:
+    if shapeenv:
+      return (" Partial solution: " +
+              ", ".join([f"{var} = {val}" for var, val in shapeenv.items()]) + ".")
+    else:
+      return ""
+
+  def process_one_eqn(eqn: DimEquation) -> bool:
+    # Try to rewrite the equation as "var * factor_var = tf_expr" (a linear
+    # uni-variate equation. Return False if this rewrite fails.
+    # Otherwise, add the variable to shapeenv and return True.
+
+    # The invariant is: var * factor_var + rest_eqn_poly = tf_expr
+    var, factor_var = None, None
+    tf_expr = eqn.tf_expr
+
+    for mon, factor in eqn.poly.items():
+      # Perhaps we can already evaluate this monomial (all vars solved)
+      try:
+        mon_value = mon.evaluate(shapeenv)
+        tf_expr = tf.math.subtract(tf_expr, _multiply(factor, mon_value))
+        continue
+      except KeyError:
+        # There are some indeterminate variables. We handle only the case of
+        # linear remaining indeterminates.
+        v = mon.to_var()
+        if v is not None and var is None:
+          var = v
+          factor_var = factor
+          continue
+      return False
+
+    if var is not None:
+      try:
+        var_value = tf.math.floordiv(tf_expr, factor_var) if factor_var != 1 else tf_expr
+        # Check that the division is even. Works only in eager mode.
+        if tf.math.floormod(tf_expr, factor_var).numpy() != 0:
+          msg = (f"Dimension variable {var} must have integer value >= 1. "  # type: ignore
+                 f"Found value {int(tf_expr.numpy()) / factor_var} when solving "
+                 f"{eqn.poly} == {eqn.tf_expr}.{_shapeenv_to_str()}")
+          raise ValueError(msg)
+        if var_value.numpy() <= 0:
+          msg = (f"Dimension variable {var} must have integer value >= 1. "
+                 f"Found value {int(var_value.numpy())} when solving "
+                 f"{eqn.poly} == {eqn.tf_expr}.{_shapeenv_to_str()}")
+          raise ValueError(msg)
+      except AttributeError:
+        # TODO(necula): check even in graph mode, by embedding the checks in
+        # the graph.
+        pass
+      shapeenv[var] = var_value
+      return True
+    else:
+      # All variables are resolved for this equation
+      try:
+        if tf_expr.numpy() != 0:
+          err_msg = (
+              "Found inconsistency when solving "
+              f"{eqn.poly} == {eqn.tf_expr}.{_shapeenv_to_str()}")
+          raise ValueError(err_msg)
+      except AttributeError:
+        # TODO(necula): check that the equation is satisfied even in graph
+        # mode.
+        pass
+      return True
+
+  while True:
+    nr_eqns = len(eqns)
+    eqns = [eqn for eqn in eqns if not process_one_eqn(eqn)]
+    if not eqns:
+      return shapeenv  # SUCCESS
+    elif len(eqns) >= nr_eqns:
+      break
+
+  # We have some equations that we cannot solve further
+  unsolved_vars: Set[str] = set()
+  unsolved_polys: List[_DimPolynomial] = []
+  for eqn in eqns:
+    unsolved_vars = unsolved_vars.union(eqn.poly.get_vars())
+    unsolved_polys.append(eqn.poly)
+  unsolved_vars = unsolved_vars.difference(shapeenv.keys())
+  eqns_str = "\n  ".join([str(eqn.poly) for eqn in eqns])
+  err_msg = (
+      f"Cannot solve for values of dimension variables {unsolved_vars} from "
+      f"the remaining dimension polynomials\n  {eqns_str}.{_shapeenv_to_str()} "
+      "Dimension variables can be solved only from linear polynomials.")
+  raise ValueError(err_msg)
