@@ -11,17 +11,17 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-"""See primitives_test docstring for how the Jax2TfLimitations are used"""
+"""See primitives_test docstring for how the Jax2TfLimitations are used."""
 
 import itertools
-import numpy as np
 from typing import Any, Callable, Optional, Sequence, Union
 
-from jax._src import dtypes
 from jax import lax
 from jax import numpy as jnp
-
+from jax import test_util as jtu
+from jax._src import dtypes
 from jax.experimental.jax2tf.tests import primitive_harness
+import numpy as np
 
 DType = Any
 
@@ -83,7 +83,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   def get_max_tolerance_limitation(
       self, limitations: Sequence["Jax2TfLimitation"]
   ) -> Optional["Jax2TfLimitation"]:
-    """Pick the tolerance limitation that establishes the maximum tolerance"""
+    """Pick the tolerance limitation that establishes the maximum tolerance."""
     # TODO: it would be best if the limitations with tolerance are mutually exclusive
     # and we don't have to compute the maximum
     # TODO: we made this an instance method only so that we don't have to import
@@ -124,15 +124,20 @@ class Jax2TfLimitation(primitive_harness.Limitation):
 
   # We keep here the explicit set of groups for which we don't have limitations
   harness_groups_no_limitations = {
-      "abs", "add", "add_any", "and", "argmin", "argmax", "atan2",
-      "bitcast_convert_type", "broadcast", "broadcast_in_dim", "ceil", "clamp",
-      "concatenate", "cos", "cosh", "complex", "conj", "convert_element_type",
+      "abs", "add", "add_any", "and", "atan2",
+      "bitcast_convert_type", "broadcast", "broadcast_in_dim", "cbrt", "ceil",
+      "clamp", "concatenate", "cos", "cosh", "complex", "conj",
+      "convert_element_type",
       "cummax", "cummin", "device_put", "dynamic_slice",
-      "dynamic_update_slice", "exp", "eq", "floor", "gather", "ge", "gt", "imag",
+      "dynamic_update_slice", "exp", "eq", "floor", "gather", "ge", "gt",
+      "imag",
       "iota", "is_finite", "le", "lt", "log", "mul", "ne", "neg", "not",
-      "or", "pad", "population_count", "random_split",
+      "or", "pad", "population_count",
+      "random_categorical", "random_split", "random_uniform", "random_randint",
+      "reduce",
       "reduce_and", "reduce_prod", "reduce_or", "reduce_sum",
-      "reduce_window_add", "reduce_window_mul", "reduce_window_min", "reduce_window_max",
+      "reduce_window_add", "reduce_window_mul", "reduce_window_min",
+      "reduce_window_max",
       "real", "reshape", "rev", "rsqrt", "scatter_max", "scatter_min",
       "select", "select_and_scatter_add",
       "shift_left", "shift_right_logical", "shift_right_arithmetic", "sign",
@@ -176,6 +181,23 @@ class Jax2TfLimitation(primitive_harness.Limitation):
         cls.helper_get_trig_custom_limitation(np.cosh)
     ]
 
+  @classmethod
+  def argmax(cls, harness: primitive_harness.Harness):
+    return [
+        Jax2TfLimitation(
+            "different results when the input contains NaN and enable_xla=False",
+            dtypes=jtu.dtypes.all_inexact,
+            devices=("cpu", "gpu", "tpu"),
+            modes=("eager", "graph", "compiled"),
+            expect_tf_error=False,
+            skip_comparison=True,
+            enabled=("nan_" in harness.name and not harness.params["enable_xla"])),
+    ]
+
+  @classmethod
+  def argmin(cls, harness: primitive_harness.Harness):
+    return cls.argmax(harness)
+
   @classmethod
   def asin(cls, harness: primitive_harness.Harness):
     return [
@@ -232,18 +254,6 @@ class Jax2TfLimitation(primitive_harness.Limitation):
           jnp.tril(result_jax), result_tf, atol=tol, err_msg=err_msg)
 
     return [
-        # See https://github.com/google/jax/pull/3775#issuecomment-659407824;
-        Jax2TfLimitation(
-            "function not compilable",
-            dtypes=[np.complex64, np.complex128],
-            devices=("cpu", "gpu"),
-            modes="compiled"),
-        missing_tf_kernel(
-            # Interesting: on TPU, complex64 works in eager
-            # mode, but fails otherwise.
-            dtypes=[np.complex64, np.complex128],
-            devices="tpu",
-            modes=("graph", "compiled")),
         # TODO: very high tolerance
         custom_numeric(
             dtypes=[np.float32, np.complex64],
@@ -395,7 +405,11 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             devices="gpu",
             modes=("eager", "graph", "compiled"),
             enabled=(harness.params["preferred_element_type"] is not None),
-            skip_comparison=True)
+            skip_comparison=True),
+        # JAX performs float16 matmuls in float32 on CPU, so the JAX result
+        # may be more precise.
+        custom_numeric(dtypes=[np.float16], devices=["cpu"], tol=1e-2,
+                       modes=("eager", "graph", "compiled")),
     ]
 
   @classmethod
@@ -518,14 +532,6 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       check_right_eigenvectors(operand, all_w_tf, all_vr_tf)
 
     return [
-        # See https://github.com/google/jax/pull/3775#issuecomment-659407824;
-        # TODO(b/181414529): enable after XLA/GPU bug is fixed.
-        Jax2TfLimitation(
-            "XLA lowering bug",
-            dtypes=(np.complex64, np.complex128),
-            devices=("gpu",),
-            modes="compiled",
-            skip_tf_run=True),
         missing_tf_kernel(
             dtypes=dtypes.bfloat16,
             devices="tpu",
@@ -804,7 +810,6 @@ class Jax2TfLimitation(primitive_harness.Limitation):
           err_msg=err_msg)
 
     return [
-        missing_tf_kernel(dtypes=[np.complex64], devices="tpu"),
         custom_numeric(
             dtypes=[np.float32, np.complex64], devices="tpu", tol=0.1),
         custom_numeric(
@@ -918,6 +923,10 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             enabled=("singularity" in harness.name)),
     ]
 
+  @classmethod
+  def rng_bit_generator(cls, harness: primitive_harness.Harness):
+    return []
+
   @classmethod
   def round(cls, harness: primitive_harness.Harness):
     return [
@@ -929,21 +938,11 @@ class Jax2TfLimitation(primitive_harness.Limitation):
 
   @classmethod
   def scatter_add(cls, harness):
-    return [
-        missing_tf_kernel(
-            dtypes=[np.complex64],
-            devices="tpu",
-        )
-    ]
+    return []
 
   @classmethod
   def scatter_mul(cls, harness):
-    return [
-        missing_tf_kernel(
-            dtypes=[np.complex64],
-            devices="tpu",
-        ),
-    ]
+    return []
 
   @classmethod
   def select_and_gather_add(cls, harness):
@@ -1059,10 +1058,6 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             first_arr_jax[~mask_jax], first_arr_tf[~mask_tf], err_msg=err_msg)
 
     return [
-        missing_tf_kernel(
-            dtypes=[np.uint64, np.int64],
-            devices=("cpu", "gpu"),
-            modes="compiled"),
         custom_numeric(
             dtypes=[np.float16, dtypes.bfloat16, np.float32, np.float64],
             custom_assert=custom_assert,
@@ -1082,6 +1077,9 @@ class Jax2TfLimitation(primitive_harness.Limitation):
         custom_numeric(dtypes=np.float32, tol=5e-3)
     ]
 
+  @classmethod
+  def tridiagonal_solve(cls, harness: primitive_harness.Harness):
+    return []
 
 def custom_numeric(
     *,
