@@ -341,7 +341,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     d_dx_jax = grad_g(x)
     d_dx_tf = jax2tf.convert(grad_g)(x)
     self.assertEqual(d_dx_jax.dtype, dtypes.float0)
-    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), dtypes.bfloat16),
+    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), np.int32),
                         d_dx_tf.numpy())
 
     shape = (3, 4)
@@ -349,32 +349,201 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     d_dx_jax = grad_g(x)
     d_dx_tf = jax2tf.convert(grad_g)(x)
     self.assertEqual(d_dx_jax.dtype, dtypes.float0)
-    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), dtypes.bfloat16),
+    self.assertAllClose(jnp.zeros(np.shape(d_dx_jax), np.int32),
                         d_dx_tf.numpy())
 
+  @parameterized.named_parameters(jtu.cases_from_list(
+      dict(testcase_name=f"function={with_function}",
+           with_function=with_function)
+      for with_function in [False, True]))
+  def test_gradients_unused_argument_readme(self, with_function=True):
+    # x2 and x3 are not used. x3 has integer type.
+    def fn(x0, x1, x2, x3):
+      return x0 * 0. + x2 * 2.
 
-  def test_tf_gradients_int_argument(self):
+    xs = [tf.Variable(x) for x in [10., 11., 12., 13]]
+    with tf.GradientTape(persistent=True) as tape:
+      res = fn(*xs)
+
+    g_tf_native = tape.gradient(res, xs)
+    self.assertAllClose(g_tf_native[0].numpy(), np.float32(0.))
+    self.assertIsNone(g_tf_native[1])
+    self.assertAllClose(g_tf_native[2].numpy(), np.float32(2.))
+    self.assertIsNone(g_tf_native[3])
+
+    g_tf_native_0 = tape.gradient(res, xs,
+                                  unconnected_gradients=tf.UnconnectedGradients.ZERO)
+    self.assertAllClose(g_tf_native_0[0].numpy(), np.float32(0.))
+    self.assertAllClose(g_tf_native_0[1].numpy(), np.float32(0.))
+    self.assertAllClose(g_tf_native_0[2].numpy(), np.float32(2.))
+    self.assertAllClose(g_tf_native_0[3].numpy(), np.int32(0))
+
+    # Now with jax2tf.convert
+    with tf.GradientTape(persistent=True) as tape:
+      conv_fn = jax2tf.convert(fn, with_gradient=True)
+      if with_function:
+        conv_fn = tf.function(conv_fn, autograph=False)
+      res = conv_fn(*xs)
+
+    g_jax2tf = tape.gradient(res, xs)
+    # Returns: 0., 0., 2., None
+    # Note that the gradient for x1 is 0.
+    self.assertAllClose(g_jax2tf[0].numpy(), np.float32(0.))
+    self.assertAllClose(g_jax2tf[1].numpy(), np.float32(0.))
+    self.assertAllClose(g_jax2tf[2].numpy(), np.float32(2.))
+    self.assertIsNone(g_jax2tf[3])
+
+    g_jax2tf = tape.gradient(res, xs,
+                               unconnected_gradients=tf.UnconnectedGradients.ZERO)
+    self.assertAllClose(g_jax2tf[0].numpy(), np.float32(0.))
+    self.assertAllClose(g_jax2tf[1].numpy(), np.float32(0.))
+    self.assertAllClose(g_jax2tf[2].numpy(), np.float32(2.))
+    self.assertAllClose(g_jax2tf[3].numpy(), np.int32(0))
+
+  # @parameterized.named_parameters(jtu.cases_from_list(
+  #     dict(testcase_name=f"function={with_function}",
+  #          with_function=with_function)
+  #     for with_function in [False, True]))
+  def test_gradients_int_argument(self, with_function=True):
     # https://github.com/google/jax/issues/6975
-    # state is a pytree that contains an integer and a boolean.
-    # The function returns an integer and a boolean.
-    def f_jax(param, state, x):
-      return param * x, state
+    # Also issue #6975.
+    # An expanded version of test_gradients_unused_argument
+    state = dict(
+        float_used=np.array([0.7, 0.9], dtype=np.float32),
+        float_passthrough=np.float16(1.),
+        float_unused=np.array([1.1, 2.2, 3.3], dtype=np.float32),
+        int_used=np.int16(5),
+        int_passthrough=np.int8(7),
+        int_unused=np.array([1, 2, 3], dtype=np.uint32),
+        bool_used=np.array([True, False, False, True], dtype=np.bool_),
+        bool_passthrough=np.array([True, False, False, True, False], dtype=np.bool_),
+        bool_unused=np.array([[True, False], [False, True]], dtype=np.bool_),
+    )
+    def jax_f(state):
+      res = dict(state,
+                 float_used=2. * state["float_used"],
+                 int_used=3 * state["int_used"],
+                 bool_used=(state["bool_used"] == state["bool_used"]))
+      del res["float_unused"]
+      del res["int_unused"]
+      del res["bool_unused"]
+      return res
 
-    param = np.array([0.7, 0.9], dtype=np.float32)
-    state = dict(array=1., counter=7, truth=True)
-    x = 3.
+    args = (state,)
+    res_jax = jax_f(*args)
+    # Native JAX AD
+    vjp_jax_fun, args_vjp = tf_test_util.TransformJaxVJP(jax_f, args, res_jax)
+    grad_jax, = vjp_jax_fun(*args_vjp)
 
-    # tf.function is important, without it the bug does not appear
-    f_tf = tf.function(jax2tf.convert(f_jax, with_gradient=True), autograph=False)
+    def compare_with_overrides(*, what, expected, **expected_overrides):
+      what_keys = set(what.keys())
+      expected_keys = set(expected.keys())
+      self.assertEqual(what_keys, expected_keys)
+      for k, w in what.items():
+        e = expected[k]
+        if k in expected_overrides:
+          if expected_overrides[k] == "ZERO":
+            e = np.zeros_like(w)
+          elif expected_overrides[k] == "ZERO_INT32":
+            e = np.zeros(np.shape(w), dtype=np.int32)
+          elif expected_overrides[k] == "ONE":
+            e = np.ones_like(w)
+          else:
+            e = expected_overrides[k]
 
-    paramv = tf.Variable(param)
-    with tf.GradientTape() as tape:
-      r, _ = f_tf(paramv, state, x)
-      loss = tf.reduce_sum(r)
+        if e is None:
+          self.assertIsNone(w, msg=k)
+        else:
+          self.assertIsNotNone(w, msg=k)
+        w = w.numpy() if isinstance(w, tf.Tensor) else e
+        e = e.numpy() if isinstance(e, tf.Tensor) else e
+        try:
+          self.assertAllClose(e, w, err_msg=k)
+        except:
+          print(f"Failed at {k}")
+          raise
+
+
+    # compare_with_overrides(g_jax, {},
+    #   bool_passthrough=np.zeros(state["bool_passthrough"].shape, dtype=dtypes.float0),
+    #   bool_unused=np.zeros(state["bool_unused"].shape, dtype=dtypes.float0),
+    #   bool_used=np.zeros(state["bool_used"].shape, dtype=dtypes.float0),
+    #   float_passthrough=np.ones_like(state["float_passthrough"]),
+    #   float_unused=np.zeros_like(state["float_unused"]),
+    #   float_used=np.ones_like(state["float_used"]) * np.array(2., dtype=state["float_used"].dtype),
+    #   int_passthrough=np.zeros(state["int_passthrough"].shape, dtype=dtypes.float0),
+    #   int_unused=np.zeros(state["int_unused"].shape, dtype=dtypes.float0),
+    #   int_used=np.zeros(state["int_used"].shape, dtype=dtypes.float0))
+
+
+    # Now native TF gradients, only to test how native TF AD works
+    _, (grad_tf_0,) = tf_test_util.ComputeTfValueAndGrad(
+        jax_f, args, unconnected_gradients=tf.UnconnectedGradients.ZERO)
+    compare_with_overrides(what=grad_tf_0,
+                           expected=grad_jax,
+                           float_unused="ZERO",
+                           bool_used="ZERO", bool_passthrough="ONE", bool_unused="ZERO",
+                           int_used="ZERO", int_passthrough="ONE", int_unused="ZERO")
+
+    _, (grad_tf_None,) = tf_test_util.ComputeTfValueAndGrad(
+        jax_f, args,
+        unconnected_gradients=tf.UnconnectedGradients.NONE)
+    compare_with_overrides(what=grad_tf_None,
+                           expected=grad_tf_0,
+                           float_unused=None, int_used=None, int_unused=None,
+                           bool_used=None, bool_unused=None)
+
+    f_tf_jax = jax2tf.convert(jax_f)
+    if with_function:
+      f_tf_jax = tf.function(f_tf_jax, autograph=False)
+
+    _, (grad_tf_jax_0,) = tf_test_util.ComputeTfValueAndGrad(f_tf_jax, args)
+    # Same results as TF native AD with tf.UnconnectedGradients.ZERO
+    compare_with_overrides(what=grad_tf_jax_0,
+                           expected=grad_tf_0,
+                           int_passthrough="ZERO", bool_passthrough="ZERO")
+
+    _, (grad_tf_jax_None,) = tf_test_util.ComputeTfValueAndGrad(
+        f_tf_jax, args,
+        unconnected_gradients=tf.UnconnectedGradients.NONE)
+    compare_with_overrides(what=grad_tf_jax_None,
+                           expected=grad_tf_0,
+                           int_used=None, int_passthrough=None, int_unused=None,
+                           bool_unused=None, bool_used=None, bool_passthrough=None)
+
+    # Not convert the JAX gradient function
+    tf_vjp_jax_fun = jax2tf.convert(vjp_jax_fun)
+    grad_tf_vjp_jax, = tf_vjp_jax_fun(*args_vjp)
+    compare_with_overrides(what=grad_tf_vjp_jax,
+                           expected=grad_tf_0,
+                           bool_passthrough="ZERO_INT32",
+                           bool_unused="ZERO_INT32", bool_used="ZERO_INT32",
+                           int_passthrough="ZERO_INT32", int_unused="ZERO_INT32",
+                           int_used="ZERO_INT32")
+
+  def test_readme_gradient_int(self):
+    x = np.array(2, dtype=np.int16)
+
+    def f_jax(x):  # x: int16
+      return x.astype(np.float32) * 2.
+
+    print(jax.grad(f_jax, allow_int=True)(x))
+    # returns a special `float0`: array((b'',), dtype=[('float0', 'V')])
+
+    print(jax2tf.convert(jax.grad(f_jax, allow_int=True))(x))
+    # returns a 0 with same shape as x, but with dtype int32
+
+    def f_tf(x):  # x: int16
+      return tf.cast(x, tf.float32) * 2.
+
+    xv = tf.Variable(x)
+    with tf.GradientTape(persistent=True) as tape:
+      print(tape.gradient(f_tf(xv), xv))
+      # returns None
+      print(tape.gradient(f_tf(xv), xv,
+                          unconnected_gradients=tf.UnconnectedGradients.ZERO))
+      # returns 0 with the same shape and dtype as x
 
-    g_tf = tape.gradient(loss, paramv)
-    self.assertAllClose(g_tf.numpy(),
-                        jax.grad(lambda *args: jnp.sum(f_jax(*args)[0]))(param, state, x))
 
   def test_convert_argument_non_callable_error(self):
     with self.assertRaisesRegex(TypeError, "Expected a callable value"):
@@ -576,15 +745,51 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         tf_fn_array(np.array([3, 4, 5])), np.array([4.5, 10, 17.5],
                                                    jnp.bfloat16))
 
-  def test_kwargs(self):
+  def test_weak_types(self):
+    mul = jax.jit(jnp.multiply)
+    # The value `2` here should be weakly typed, and should not lead to
+    # promotion.
+    tf_fn = jax2tf.convert(lambda x: mul(x, 2.))
+    self.assertAllClose(tf_fn(tf.constant(1.375, tf.bfloat16)).numpy(),
+                        jnp.bfloat16(2.750))
+
+  @parameterized.named_parameters(jtu.cases_from_list(
+      dict(testcase_name=f"function={with_function}",
+           with_function=with_function)
+      for with_function in [False, True]))
+  def test_kwargs(self, with_function=True):
     # Re: https://github.com/google/jax/issues/6791
     def f_jax(*, x):
       return jnp.sum(x)
     f_tf = jax2tf.convert(f_jax)
+    if with_function:
+      f_tf = tf.function(f_tf)
     self.assertAllClose(
       f_tf(x=np.zeros(3, dtype=np.float32)),  # Call with kwargs.
       np.zeros((), dtype=np.float32))
 
+  @parameterized.named_parameters(jtu.cases_from_list(
+      dict(testcase_name=f"function={with_function}",
+           with_function=with_function)
+      for with_function in [False, True]))
+  def test_grad_kwargs(self, with_function=False):
+    # Re: https://github.com/google/jax/issues/6791
+    x = (np.zeros(3, dtype=np.float32),
+         np.zeros(4, dtype=np.float32))
+    def f_jax(*, x=(1., 2.)):
+      return jnp.sum(x[0]) + 2. * jnp.sum(x[1])
+    f_tf = jax2tf.convert(f_jax)
+    if with_function:
+      f_tf = tf.function(f_tf)
+    xv = tf.nest.map_structure(tf.Variable, x)
+    with tf.GradientTape() as tape:
+      res = f_tf(x=xv)
+    grad_tf = tape.gradient(res, xv)
+    self.assertAllClose((np.full_like(x[0], fill_value=1.),
+                         np.full_like(x[1], fill_value=2.)),
+                        (grad_tf[0].numpy(), grad_tf[1].numpy()))
+
+
   def test_enable_xla(self):
     # Tests that enable_xla flag is properly scoped to a conversion.
     def fun(x):
