@@ -21,7 +21,6 @@ import itertools as it
 from functools import partial
 
 from . import maps
-from . import PartitionSpec
 from .. import core
 from .. import linear_util as lu
 from .._src.api import _check_callable, _check_arg
@@ -35,6 +34,7 @@ from ..interpreters import pxla
 from ..interpreters import xla
 from ..interpreters import batching
 from ..interpreters import partial_eval as pe
+from ..interpreters.sharded_jit import PartitionSpec
 from ..lib import xla_bridge as xb
 from ..lib import xla_client as xc
 from ..tree_util import tree_flatten, tree_unflatten
@@ -56,9 +56,9 @@ def pjit(fun: Callable,
   version of ``fun`` would not fit in a single device's memory, or to speed up
   ``fun`` by running each operation in parallel across multiple devices.
 
-  The partitioning over devices happens automatically based on
-  propagation of input partitioning specified in ``in_axis_resources`` and
-  output partitioning specified in ``out_axis_resources``. The resources
+  The partitioning over devices happens automatically based on the
+  propagation of the input partitioning specified in ``in_axis_resources`` and
+  the output partitioning specified in ``out_axis_resources``. The resources
   specified in those two arguments must refer to mesh axes, as defined by
   the :py:func:`jax.experimental.maps.mesh` context manager. Note that the mesh
   definition at ``pjit`` application time is ignored, and the returned function
@@ -84,7 +84,7 @@ def pjit(fun: Callable,
     mesh axis size, and outputs will be similarly sized according to the local
     mesh. ``fun`` will still be executed across *all* devices in the mesh,
     including those from other processes, and will be given a global view of the
-    data spread accross multiple processes as a single array. However, outside
+    data spread across multiple processes as a single array. However, outside
     of ``pjit`` every process only "sees" its local piece of the input and output,
     corresponding to its local sub-mesh.
 
@@ -140,7 +140,7 @@ def pjit(fun: Callable,
 
   Returns:
     A wrapped version of ``fun``, set up for just-in-time compilation and
-    automatic partitioned by the mesh available at each call site.
+    automaticly partitioned by the mesh available at each call site.
 
   For example, a convolution operator can be automatically partitioned over
   an arbitrary set of devices by a single ```pjit`` application:
@@ -265,6 +265,7 @@ class SpecSync(IntEnum):
   We use this to make sure we don't show garbage modified values while claiming
   that the users have specified them like that.
   """
+  OUT_OF_SYNC = 0  # Arbitrary changes, including new axes inserted
   DIM_PERMUTE = 1  # Dimensions permuted, but no new sharding axes
   IN_SYNC = 2  # Entirely in sync
 
@@ -291,7 +292,7 @@ class ParsedPartitionSpec:
     if too_short > 0:
       parts += ((),) * too_short
     new_partitions = tuple_insert(parts, dim, val)
-    new_sync = SpecSync.DIM_PERMUTE if val == () else SpecSync.IN_SYNC
+    new_sync = SpecSync.DIM_PERMUTE if val == () else SpecSync.OUT_OF_SYNC
     return ParsedPartitionSpec(self.unsafe_user_spec, new_partitions, sync=new_sync)
 
   @classmethod
@@ -329,6 +330,10 @@ class ParsedPartitionSpec:
   def __iter__(self):
     return iter(self.partitions)
 
+  def __repr__(self):
+    return f"<partitions={self.partitions} sync={self.sync}>"
+
+
 REPLICATED = ParsedPartitionSpec(None, ())
 
 
@@ -566,20 +571,33 @@ def _sharding_constraint_impl(x, axis_resources, resource_env):
   raise NotImplementedError(
       "with_sharding_constraint() should only be called inside pjit()")
 
-def _sharding_constraint_translation_rule(c, x_node, axis_resources, resource_env):
-  mesh = resource_env.physical_mesh
-  return xb.set_sharding_proto(c, x_node,
-                               get_sharding_proto(c, x_node, axis_resources, mesh))
-
 sharding_constraint_p = core.Primitive("sharding_constraint")
 sharding_constraint_p.def_impl(_sharding_constraint_impl)
-sharding_constraint_p.def_abstract_eval(lambda x, **unused: x)
+sharding_constraint_p.def_abstract_eval(lambda x, **_: x)
 ad.deflinear2(sharding_constraint_p,
               lambda ct, _, axis_resources, resource_env: (
                   sharding_constraint_p.bind(
                       ct, axis_resources=axis_resources, resource_env=resource_env),))
+
+def _sharding_constraint_translation_rule(c, x_node, axis_resources, resource_env):
+  mesh = resource_env.physical_mesh
+  return xb.set_sharding_proto(c, x_node,
+                               get_sharding_proto(c, x_node, axis_resources, mesh))
 xla.translations[sharding_constraint_p] = _sharding_constraint_translation_rule
 
+def _sharding_constraint_batcher(insert_axis, vals_in, dims_in, axis_resources, resource_env,
+                                 axis_name, main_type):
+  x, = vals_in
+  d, = dims_in
+  new_parts = (axis_name,) if insert_axis else ()
+  y = sharding_constraint_p.bind(
+      x,
+      axis_resources=axis_resources.insert_axis_partitions(d, new_parts),
+      resource_env=resource_env)
+  return y, d
+batching.initial_style_batchers[sharding_constraint_p] = partial(_sharding_constraint_batcher, False)
+pxla.spmd_primitive_batchers[sharding_constraint_p] = partial(_sharding_constraint_batcher, True)
+
 def _resource_typing_sharding_constraint(avals, params, source_info, named_axis_resources):
   aval, = avals
   _check_resources_against_named_axes(
