@@ -17,6 +17,8 @@ from .ad import grad, value_and_grad
 from .ops import (
     bcoo_dot_general,
     bcoo_dot_general_p,
+    bcoo_dot_general_sampled,
+    bcoo_dot_general_sampled_p,
     bcoo_extract,
     bcoo_extract_p,
     bcoo_fromdense,
@@ -48,3 +50,5 @@ from .ops import (
     CSR,
     BCOO,
 )
+
+from .transform import sparsify
\ No newline at end of file
