@@ -23,8 +23,8 @@ product, sparse matrix/matrix product) for two common sparse representations
 
 These routines have reference implementations defined via XLA scatter/gather
 operations that will work on any backend, although they are not particularly
-performant. On GPU runtimes with jaxlib 0.1.66 or newer built against CUDA 11.0
-or newer, each operation is computed efficiently via cusparse.
+performant. On GPU runtimes built against CUDA 11.0 or newer, each operation is
+computed efficiently via cusparse.
 
 Further down are some examples of potential high-level wrappers for sparse objects.
 (API should be considered unstable and subject to change).
@@ -36,11 +36,13 @@ from typing import Any, Sequence, Tuple
 
 from jax import api
 from jax import core
+from jax import dtypes
 from jax import jit
 from jax import lax
 from jax import tree_util
 from jax import vmap
 from jax.interpreters import batching
+from jax.interpreters import partial_eval as pe
 from jax.interpreters import xla
 from jax.lib import cusparse
 from jax.lib import xla_bridge
@@ -49,7 +51,7 @@ import jax.numpy as jnp
 import numpy as np
 from jax.interpreters import ad
 from jax.util import safe_zip
-from jax._src.lax.lax import ranges_like, remaining, _dot_general_batch_dim_nums
+from jax._src.lax.lax import ranges_like, remaining, _dot_general_batch_dim_nums, _dot_general_shape_computation
 
 xb = xla_bridge
 xops = xla_client.ops
@@ -60,8 +62,8 @@ Dtype = Any
 # TODO: possibly make these utilities into primitives, targeting
 #       csr2coo/coo2csr/SPDDMM
 @functools.partial(jit, static_argnums=1)
-def _csr_to_coo(indptr, nnz):
-  return jnp.cumsum(jnp.zeros_like(indptr, shape=nnz).at[indptr].add(1)) - 1
+def _csr_to_coo(indptr, nse):
+  return jnp.cumsum(jnp.zeros_like(indptr, shape=nse).at[indptr].add(1)) - 1
 
 @functools.partial(jit, static_argnums=1)
 def _coo_to_csr(row, nrows):
@@ -87,8 +89,8 @@ def csr_todense(data, indices, indptr, *, shape):
   """Convert CSR-format sparse matrix to a dense matrix.
 
   Args:
-    data : array of shape ``(nnz,)``.
-    indices : array of shape ``(nnz,)``
+    data : array of shape ``(nse,)``.
+    indices : array of shape ``(nse,)``
     indptr : array of shape ``(shape[0] + 1,)`` and dtype ``indices.dtype``
     shape : length-2 tuple representing the matrix shape
 
@@ -124,33 +126,33 @@ if cusparse and cusparse.is_supported:
 csr_fromdense_p = core.Primitive('csr_fromdense')
 csr_fromdense_p.multiple_results = True
 
-def csr_fromdense(mat, *, nnz, index_dtype=np.int32):
+def csr_fromdense(mat, *, nse, index_dtype=np.int32):
   """Create CSR-format sparse matrix from a dense matrix.
 
   Args:
     mat : array to be converted to CSR.
-    nnz : number of nonzero entries in ``mat``
+    nse : number of specified entries in ``mat``
     index_dtype : dtype of sparse indices
 
   Returns:
-    data : array of shape ``(nnz,)`` and dtype ``mat.dtype``.
-    indices : array of shape ``(nnz,)`` and dtype ``index_dtype``
+    data : array of shape ``(nse,)`` and dtype ``mat.dtype``.
+    indices : array of shape ``(nse,)`` and dtype ``index_dtype``
     indptr : array of shape ``(mat.shape[0] + 1,)`` and dtype ``index_dtype``
   """
   mat = jnp.asarray(mat)
-  nnz = core.concrete_or_error(operator.index, nnz, "nnz argument of csr_fromdense()")
-  return csr_fromdense_p.bind(mat, nnz=nnz, index_dtype=np.dtype(index_dtype))
+  nse = core.concrete_or_error(operator.index, nse, "nse argument of csr_fromdense()")
+  return csr_fromdense_p.bind(mat, nse=nse, index_dtype=np.dtype(index_dtype))
 
 @csr_fromdense_p.def_impl
-def _csr_fromdense_impl(mat, *, nnz, index_dtype):
+def _csr_fromdense_impl(mat, *, nse, index_dtype):
   mat = jnp.asarray(mat)
   assert mat.ndim == 2
   m = mat.shape[0]
 
-  row, col = jnp.nonzero(mat, size=nnz)
+  row, col = jnp.nonzero(mat, size=nse)
   data = mat[row, col]
 
-  true_nonzeros = jnp.arange(nnz) < (mat != 0).sum()
+  true_nonzeros = jnp.arange(nse) < (mat != 0).sum()
   data = jnp.where(true_nonzeros, data, 0)
   row = jnp.where(true_nonzeros, row, m)
   indices = col.astype(index_dtype)
@@ -159,15 +161,15 @@ def _csr_fromdense_impl(mat, *, nnz, index_dtype):
   return data, indices, indptr
 
 @csr_fromdense_p.def_abstract_eval
-def _csr_fromdense_abstract_eval(mat, *, nnz, index_dtype):
-  data = core.ShapedArray((nnz,), mat.dtype)
-  indices = core.ShapedArray((nnz,), index_dtype)
+def _csr_fromdense_abstract_eval(mat, *, nse, index_dtype):
+  data = core.ShapedArray((nse,), mat.dtype)
+  indices = core.ShapedArray((nse,), index_dtype)
   indptr = core.ShapedArray((mat.shape[0] + 1,), index_dtype)
   return data, indices, indptr
 
-def _csr_fromdense_gpu_translation_rule(c, mat, *, nnz, index_dtype):
+def _csr_fromdense_gpu_translation_rule(c, mat, *, nse, index_dtype):
   data, indices, indptr = cusparse.csr_fromdense(
-      c, mat, nnz=nnz, index_dtype=np.dtype(index_dtype))
+      c, mat, nnz=nse, index_dtype=np.dtype(index_dtype))
   return xops.Tuple(c, [data, indices, indptr])
 
 xla.translations[csr_fromdense_p] = xla.lower_fun(
@@ -185,8 +187,8 @@ def csr_matvec(data, indices, indptr, v, *, shape, transpose=False):
   """Product of CSR sparse matrix and a dense vector.
 
   Args:
-    data : array of shape ``(nnz,)``.
-    indices : array of shape ``(nnz,)``
+    data : array of shape ``(nse,)``.
+    indices : array of shape ``(nse,)``
     indptr : array of shape ``(shape[0] + 1,)`` and dtype ``indices.dtype``
     v : array of shape ``(shape[0] if transpose else shape[1],)``
       and dtype ``data.dtype``
@@ -214,7 +216,7 @@ def _csr_matvec_abstract_eval(data, indices, indptr, v, *, shape, transpose):
   assert indices.dtype == indptr.dtype
   assert len(indptr) == shape[0] + 1
   out_shape = shape[1] if transpose else shape[0]
-  assert v.shape == (shape[0],) if transpose else (shape[1],)
+  assert v.shape[0] == (shape[0] if transpose else shape[1])
   return core.ShapedArray((out_shape,), data.dtype)
 
 def _csr_matvec_gpu_translation_rule(c, data, indices, indptr, v, *, shape, transpose):
@@ -236,8 +238,8 @@ def csr_matmat(data, indices, indptr, B, *, shape, transpose=False):
   """Product of CSR sparse matrix and a dense matrix.
 
   Args:
-    data : array of shape ``(nnz,)``.
-    indices : array of shape ``(nnz,)``
+    data : array of shape ``(nse,)``.
+    indices : array of shape ``(nse,)``
     indptr : array of shape ``(shape[0] + 1,)`` and dtype ``indices.dtype``
     B : array of shape ``(shape[0] if transpose else shape[1], cols)`` and
       dtype ``data.dtype``
@@ -258,6 +260,7 @@ def _csr_matmat_impl(data, indices, indptr, B, *, shape, transpose):
 
 @csr_matmat_p.def_abstract_eval
 def _csr_matmat_abstract_eval(data, indices, indptr, B, *, shape, transpose):
+  assert len(shape) == 2
   assert data.ndim == indices.ndim == indptr.ndim == 1
   assert B.ndim == 2
   assert data.shape == indices.shape
@@ -265,7 +268,7 @@ def _csr_matmat_abstract_eval(data, indices, indptr, B, *, shape, transpose):
   assert indices.dtype == indptr.dtype
   assert len(indptr) == shape[0] + 1
   out_shape = shape[1] if transpose else shape[0]
-  assert B.shape[0] == shape[0] if transpose else shape[1]
+  assert B.shape[0] == (shape[0] if transpose else shape[1])
   return core.ShapedArray((out_shape, B.shape[1]), data.dtype)
 
 def _csr_matmat_gpu_translation_rule(c, data, indices, indptr, B, *, shape, transpose):
@@ -287,9 +290,9 @@ def coo_todense(data, row, col, *, shape):
   """Convert CSR-format sparse matrix to a dense matrix.
 
   Args:
-    data : array of shape ``(nnz,)``.
-    row : array of shape ``(nnz,)``
-    col : array of shape ``(nnz,)`` and dtype ``row.dtype``
+    data : array of shape ``(nse,)``.
+    row : array of shape ``(nse,)``
+    col : array of shape ``(nse,)`` and dtype ``row.dtype``
     shape : length-2 tuple representing the matrix shape
 
   Returns:
@@ -336,52 +339,52 @@ if cusparse and cusparse.is_supported:
 coo_fromdense_p = core.Primitive('coo_fromdense')
 coo_fromdense_p.multiple_results = True
 
-def coo_fromdense(mat, *, nnz, index_dtype=jnp.int32):
+def coo_fromdense(mat, *, nse, index_dtype=jnp.int32):
   """Create COO-format sparse matrix from a dense matrix.
 
   Args:
     mat : array to be converted to COO.
-    nnz : number of nonzero entries in ``mat``
+    nse : number of specified entries in ``mat``
     index_dtype : dtype of sparse indices
 
   Returns:
-    data : array of shape ``(nnz,)`` and dtype ``mat.dtype``
-    row : array of shape ``(nnz,)`` and dtype ``index_dtype``
-    col : array of shape ``(nnz,)`` and dtype ``index_dtype``
+    data : array of shape ``(nse,)`` and dtype ``mat.dtype``
+    row : array of shape ``(nse,)`` and dtype ``index_dtype``
+    col : array of shape ``(nse,)`` and dtype ``index_dtype``
   """
   mat = jnp.asarray(mat)
-  nnz = core.concrete_or_error(operator.index, nnz, "nnz argument of coo_fromdense()")
-  return coo_fromdense_p.bind(mat, nnz=nnz, index_dtype=index_dtype)
+  nse = core.concrete_or_error(operator.index, nse, "nse argument of coo_fromdense()")
+  return coo_fromdense_p.bind(mat, nse=nse, index_dtype=index_dtype)
 
 @coo_fromdense_p.def_impl
-def _coo_fromdense_impl(mat, *, nnz, index_dtype):
+def _coo_fromdense_impl(mat, *, nse, index_dtype):
   mat = jnp.asarray(mat)
   assert mat.ndim == 2
 
-  row, col = jnp.nonzero(mat, size=nnz)
+  row, col = jnp.nonzero(mat, size=nse)
   data = mat[row, col]
 
-  true_nonzeros = jnp.arange(nnz) < (mat != 0).sum()
+  true_nonzeros = jnp.arange(nse) < (mat != 0).sum()
   data = jnp.where(true_nonzeros, data, 0)
 
   return data, row.astype(index_dtype), col.astype(index_dtype)
 
 @coo_fromdense_p.def_abstract_eval
-def _coo_fromdense_abstract_eval(mat, *, nnz, index_dtype):
-  data = core.ShapedArray((nnz,), mat.dtype)
-  row = col = core.ShapedArray((nnz,), index_dtype)
+def _coo_fromdense_abstract_eval(mat, *, nse, index_dtype):
+  data = core.ShapedArray((nse,), mat.dtype)
+  row = col = core.ShapedArray((nse,), index_dtype)
   return data, row, col
 
-def _coo_fromdense_gpu_translation_rule(c, mat, *, nnz, index_dtype):
+def _coo_fromdense_gpu_translation_rule(c, mat, *, nse, index_dtype):
   data, row, col = cusparse.coo_fromdense(
-      c, mat, nnz=nnz, index_dtype=np.dtype(index_dtype))
+      c, mat, nnz=nse, index_dtype=np.dtype(index_dtype))
   return xops.Tuple(c, [data, row, col])
 
-def _coo_fromdense_jvp(primals, tangents, *, nnz, index_dtype):
+def _coo_fromdense_jvp(primals, tangents, *, nse, index_dtype):
   M, = primals
   Mdot, = tangents
 
-  primals_out = coo_fromdense(M, nnz=nnz, index_dtype=index_dtype)
+  primals_out = coo_fromdense(M, nse=nse, index_dtype=index_dtype)
   data, row, col = primals_out
 
   if type(Mdot) is ad.Zero:
@@ -393,9 +396,9 @@ def _coo_fromdense_jvp(primals, tangents, *, nnz, index_dtype):
 
   return primals_out, tangents_out
 
-def _coo_fromdense_transpose(ct, M, *, nnz, index_dtype):
+def _coo_fromdense_transpose(ct, M, *, nse, index_dtype):
   data, row, col = ct
-  assert len(data) == nnz
+  assert len(data) == nse
   assert row.dtype == col.dtype == index_dtype
   if isinstance(row, ad.Zero) or isinstance(col, ad.Zero):
     raise ValueError("Cannot transpose with respect to sparse indices")
@@ -420,9 +423,9 @@ def coo_matvec(data, row, col, v, *, shape, transpose=False):
   """Product of COO sparse matrix and a dense vector.
 
   Args:
-    data : array of shape ``(nnz,)``.
-    row : array of shape ``(nnz,)``
-    col : array of shape ``(nnz,)`` and dtype ``row.dtype``
+    data : array of shape ``(nse,)``.
+    row : array of shape ``(nse,)``
+    col : array of shape ``(nse,)`` and dtype ``row.dtype``
     v : array of shape ``(shape[0] if transpose else shape[1],)`` and
       dtype ``data.dtype``
     shape : length-2 tuple representing the matrix shape
@@ -450,7 +453,8 @@ def _coo_matvec_abstract_eval(data, row, col, v, *, shape, transpose):
   assert data.dtype == v.dtype
   assert row.dtype == col.dtype
   assert len(shape) == 2
-  assert v.shape == (shape[0],) if transpose else (shape[1],)
+  assert v.ndim == 1
+  assert v.shape[0] == (shape[0] if transpose else shape[1])
   out_shape = shape[1] if transpose else shape[0]
   return core.ShapedArray((out_shape,), data.dtype)
 
@@ -491,9 +495,9 @@ def coo_matmat(data, row, col, B, *, shape, transpose=False):
   """Product of COO sparse matrix and a dense matrix.
 
   Args:
-    data : array of shape ``(nnz,)``.
-    row : array of shape ``(nnz,)``
-    col : array of shape ``(nnz,)`` and dtype ``row.dtype``
+    data : array of shape ``(nse,)``.
+    row : array of shape ``(nse,)``
+    col : array of shape ``(nse,)`` and dtype ``row.dtype``
     B : array of shape ``(shape[0] if transpose else shape[1], cols)`` and
       dtype ``data.dtype``
     shape : length-2 tuple representing the matrix shape
@@ -519,8 +523,9 @@ def _coo_matmat_impl(data, row, col, B, *, shape, transpose):
 def _coo_matmat_abstract_eval(data, row, col, B, *, shape, transpose):
   assert data.shape == row.shape == col.shape
   assert data.dtype == B.dtype
+  assert B.ndim == 2
   assert len(shape) == 2
-  assert B.shape[0] == shape[0] if transpose else shape[1]
+  assert B.shape[0] == (shape[0] if transpose else shape[1])
   out_shape = shape[1] if transpose else shape[0]
   return core.ShapedArray((out_shape, B.shape[1]), data.dtype)
 
@@ -848,7 +853,10 @@ bcoo_transpose_p = core.Primitive('bcoo_transpose')
 bcoo_transpose_p.multiple_results = True
 
 def bcoo_transpose(data, indices, *, permutation, shape):
-  return bcoo_transpose_p.bind(data, indices, permutation=permutation, shape=shape)
+  if tuple(permutation) == tuple(range(len(shape))):
+    return data, indices
+  else:
+    return bcoo_transpose_p.bind(data, indices, permutation=permutation, shape=shape)
 
 def _validate_permutation(data, indices, permutation, shape):
   if not isinstance(permutation, (tuple, list, np.ndarray)):
@@ -1050,14 +1058,26 @@ def _bcoo_dot_general_transpose(ct, lhs_data, lhs_indices, rhs, *, dimension_num
   rhs_ndim = rhs.aval.ndim if ad.is_undefined_primal(rhs) else rhs.ndim
   lhs_kept = remaining(range(lhs_ndim), lhs_contract, lhs_batch)
   rhs_kept = remaining(range(rhs_ndim), rhs_contract, rhs_batch)
-  ans_batch, ans_lhs, ans_rhs = ranges_like(lhs_batch, lhs_kept, rhs_kept)
+  ans_batch, ans_lhs, ans_rhs = map(list, ranges_like(lhs_batch, lhs_kept, rhs_kept))
   if ad.is_undefined_primal(lhs_data):
     dims = ((ans_rhs, rhs_kept), (ans_batch, rhs_batch))
     lhs_contract_sorted_by_rhs = list(np.take(lhs_contract, np.argsort(rhs_contract)))
-    # TODO: extract these sparse indices without constructing the dense matrix.
-    out_axes = np.argsort(list(lhs_batch) + lhs_kept + lhs_contract_sorted_by_rhs)
-    out_dense = lax.transpose(lax.dot_general(ct, rhs, dimension_numbers=dims), out_axes)
-    return bcoo_extract(lhs_indices, out_dense), lhs_indices, rhs
+    permutation = list(lhs_batch) + lhs_kept + lhs_contract_sorted_by_rhs
+    out_axes = np.argsort(permutation)
+
+    # What follows is essentially this, but computed in terms of dot_general_sampled:
+    # out_dense_T = lax.dot_general(ct, rhs, dimension_numbers=dims)
+    # out_dense = lax.transpose(out_dense_T, out_axes)
+    # result = bcoo_extract(lhs_indices, out_dense)
+
+    # Instead we (1) un-transpose indices, (2) compute SDDMM, (3) re-transpose result
+    dummy_data = jnp.ones([1 for i in range(lhs_indices.ndim - 2)] + [lhs_indices.shape[-1]])
+    dummy_shape = tuple(lhs_indices.shape[:-2]) + tuple(1 for i in range(lhs_indices.shape[-2]))
+    _, lhs_indices_T = bcoo_transpose(dummy_data, lhs_indices, permutation=permutation, shape=dummy_shape)
+    result_T = bcoo_dot_general_sampled(ct, rhs, lhs_indices_T, dimension_numbers=dims)
+    result, _ = bcoo_transpose(result_T, lhs_indices_T, permutation=out_axes, shape=dummy_shape)
+
+    return result, lhs_indices, rhs
   else:
     dims = ((lhs_kept, ans_lhs), (lhs_batch, ans_batch))
     rhs_contract_sorted_by_lhs = list(np.take(rhs_contract, np.argsort(lhs_contract)))
@@ -1091,6 +1111,59 @@ batching.primitive_batchers[bcoo_dot_general_p] = _bcoo_dot_general_batch_rule
 xla.translations[bcoo_dot_general_p] = xla.lower_fun(
     _bcoo_dot_general_impl, multiple_results=False)
 
+#----------------------------------------------------------------------
+# bcoo_dot_general_sampled
+# (batched) general sampled dot product of two dense ND arrays, with
+# output computed only at a given set of sparse indices.
+
+bcoo_dot_general_sampled_p = core.Primitive("bcoo_dot_general_sampled")
+
+def bcoo_dot_general_sampled(A, B, indices, *, dimension_numbers):
+  return bcoo_dot_general_sampled_p.bind(A, B, indices, dimension_numbers=dimension_numbers)
+
+@bcoo_dot_general_sampled_p.def_impl
+def _bcoo_dot_general_sampled_impl(A, B, indices, *, dimension_numbers):
+  # TODO(jakevdp): use a more efficient implementation that avoids the full dot product.
+  dense_result = lax.dot_general(A, B, dimension_numbers=dimension_numbers)
+  return bcoo_extract(indices, dense_result)
+
+@bcoo_dot_general_sampled_p.def_abstract_eval
+def _bcoo_dot_general_sampled_abstract_eval(A, B, indices, *, dimension_numbers):
+  dense_result, = pe.abstract_eval_fun(lambda *args: [lax.dot_general(*args, dimension_numbers=dimension_numbers)], A, B)
+  sparse_result, = pe.abstract_eval_fun(lambda *args: [bcoo_extract(*args)], indices, dense_result)
+  return sparse_result
+
+def _bcoo_dot_general_sampled_transpose(ct, A, B, indices, *, dimension_numbers):
+  A_shape = A.aval.shape if hasattr(A, 'aval') else A.shape
+  B_shape = B.aval.shape if hasattr(B, 'aval') else B.shape
+  mat_shape = _dot_general_shape_computation(
+    A_shape, B_shape, dimension_numbers=dimension_numbers)
+  mat = ad.UndefinedPrimal(core.ShapedArray(mat_shape, ct.dtype))
+  indices, ct = _bcoo_extract_transpose(ct, indices, mat)
+  kwds = {'dimension_numbers': dimension_numbers,
+          'precision': None,
+          'preferred_element_type': None}
+  A, B = ad.get_primitive_transpose(lax.dot_general_p)(ct, A, B, **kwds)
+  return A, B, indices
+
+def _bcoo_dot_general_sampled_jvp_A(A_dot, A, B, indices, *, dimension_numbers):
+  return bcoo_dot_general_sampled(A_dot, B, indices, dimension_numbers=dimension_numbers)
+
+def _bcoo_dot_general_sampled_jvp_B(B_dot, A, B, indices, *, dimension_numbers):
+  return bcoo_dot_general_sampled(A, B_dot, indices, dimension_numbers=dimension_numbers)
+
+def _bcoo_dot_general_sampled_batch_rule(batched_args, batch_dims, *, dimension_numbers):
+  def impl(A, B, indices):
+    return _bcoo_dot_general_sampled_impl(A, B, indices, dimension_numbers=dimension_numbers)
+  return vmap(impl, in_axes=batch_dims, out_axes=0)(*batched_args), 0
+
+ad.defjvp(bcoo_dot_general_sampled_p, _bcoo_dot_general_sampled_jvp_A,
+          _bcoo_dot_general_sampled_jvp_B, None)
+ad.primitive_transposes[bcoo_dot_general_sampled_p] = _bcoo_dot_general_sampled_transpose
+batching.primitive_batchers[bcoo_dot_general_sampled_p] = _bcoo_dot_general_sampled_batch_rule
+xla.translations[bcoo_dot_general_sampled_p] = xla.lower_fun(
+    _bcoo_dot_general_sampled_impl, multiple_results=False)
+
 #----------------------------------------------------------------------
 # BCOO functions that maybe should be primitives?
 
@@ -1150,7 +1223,7 @@ class JAXSparse:
   """Base class for high-level JAX sparse objects."""
   data: jnp.ndarray
   shape: Tuple[int, int]
-  nnz: property
+  nse: property
   dtype: property
 
   @property
@@ -1161,7 +1234,7 @@ class JAXSparse:
     self.shape = shape
 
   def __repr__(self):
-    repr_ = f"{self.__class__.__name__}({self.dtype}{list(self.shape)}, nnz={self.nnz})"
+    repr_ = f"{self.__class__.__name__}({self.dtype}{list(self.shape)}, nse={self.nse})"
     if isinstance(self.data, core.Tracer):
       repr_ = f"{type(self.data).__name__}[{repr_}]"
     return repr_
@@ -1204,7 +1277,7 @@ class CSR(JAXSparse):
   data: jnp.ndarray
   indices: jnp.ndarray
   indptr: jnp.ndarray
-  nnz = property(lambda self: self.data.size)
+  nse = property(lambda self: self.data.size)
   dtype = property(lambda self: self.data.dtype)
 
   def __init__(self, args, *, shape):
@@ -1212,10 +1285,10 @@ class CSR(JAXSparse):
     super().__init__(args, shape=shape)
 
   @classmethod
-  def fromdense(cls, mat, *, nnz=None, index_dtype=np.int32):
-    if nnz is None:
-      nnz = (mat != 0).sum()
-    return cls(csr_fromdense(mat, nnz=nnz, index_dtype=index_dtype), shape=mat.shape)
+  def fromdense(cls, mat, *, nse=None, index_dtype=np.int32):
+    if nse is None:
+      nse = (mat != 0).sum()
+    return cls(csr_fromdense(mat, nse=nse, index_dtype=index_dtype), shape=mat.shape)
 
   @api.jit
   def todense(self):
@@ -1243,7 +1316,7 @@ class CSC(JAXSparse):
   data: jnp.ndarray
   indices: jnp.ndarray
   indptr: jnp.ndarray
-  nnz = property(lambda self: self.data.size)
+  nse = property(lambda self: self.data.size)
   dtype = property(lambda self: self.data.dtype)
 
   def __init__(self, args, *, shape):
@@ -1251,10 +1324,10 @@ class CSC(JAXSparse):
     super().__init__(args, shape=shape)
 
   @classmethod
-  def fromdense(cls, mat, *, nnz=None, index_dtype=np.int32):
-    if nnz is None:
-      nnz = (mat != 0).sum()
-    return cls(csr_fromdense(mat.T, nnz=nnz, index_dtype=index_dtype), shape=mat.shape)
+  def fromdense(cls, mat, *, nse=None, index_dtype=np.int32):
+    if nse is None:
+      nse = (mat != 0).sum()
+    return cls(csr_fromdense(mat.T, nse=nse, index_dtype=index_dtype), shape=mat.shape)
 
   @api.jit
   def todense(self):
@@ -1282,7 +1355,7 @@ class COO(JAXSparse):
   data: jnp.ndarray
   row: jnp.ndarray
   col: jnp.ndarray
-  nnz = property(lambda self: self.data.size)
+  nse = property(lambda self: self.data.size)
   dtype = property(lambda self: self.data.dtype)
 
   def __init__(self, args, *, shape):
@@ -1290,10 +1363,10 @@ class COO(JAXSparse):
     super().__init__(args, shape=shape)
 
   @classmethod
-  def fromdense(cls, mat, *, nnz=None, index_dtype=np.int32):
-    if nnz is None:
-      nnz = (mat != 0).sum()
-    return cls(coo_fromdense(mat, nnz=nnz, index_dtype=index_dtype), shape=mat.shape)
+  def fromdense(cls, mat, *, nse=None, index_dtype=np.int32):
+    if nse is None:
+      nse = (mat != 0).sum()
+    return cls(coo_fromdense(mat, nse=nse, index_dtype=index_dtype), shape=mat.shape)
 
   @api.jit
   def todense(self):
@@ -1315,16 +1388,20 @@ class COO(JAXSparse):
     return (self.data, self.row, self.col), {"shape": self.shape}
 
 
-def _is_dummy(*args):
+def _is_placeholder(*args):
   return all(type(arg) is object for arg in args) or all(arg is None for arg in args)
 
+def _asarray_or_float0(arg):
+  if isinstance(arg, np.ndarray) and arg.dtype == dtypes.float0:
+    return arg
+  return jnp.asarray(arg)
 
 @tree_util.register_pytree_node_class
 class BCOO(JAXSparse):
   """Experimental BCOO matrix implemented in JAX; API subject to change."""
   data: jnp.ndarray
   indices: jnp.ndarray
-  nnz = property(lambda self: self.data.size)
+  nse = property(lambda self: self.data.size)
   dtype = property(lambda self: self.data.dtype)
   n_batch = property(lambda self: self.indices.ndim - 2)
   n_sparse = property(lambda self: self.indices.shape[-2])
@@ -1336,12 +1413,23 @@ class BCOO(JAXSparse):
     return tuple(self.shape[self.indices.ndim - 2:][:self.indices.shape[-2]])
 
   def __init__(self, args, *, shape):
-    self.data, self.indices = args
+    # JAX transforms will sometimes instantiate pytrees with null values, so we
+    # must catch that in the initialization of inputs.
+    self.data, self.indices = args if _is_placeholder(*args) else map(_asarray_or_float0, args)
     super().__init__(args, shape=shape)
 
   @classmethod
-  def fromdense(cls, mat, *, nnz=None, index_dtype=np.int32, n_dense=0, n_batch=0):
-    return cls(bcoo_fromdense(mat, nse=nnz, index_dtype=index_dtype, n_dense=n_dense, n_batch=n_batch), shape=mat.shape)
+  def fromdense(cls, mat, *, nse=None, index_dtype=np.int32, n_dense=0, n_batch=0):
+    return cls(bcoo_fromdense(mat, nse=nse, index_dtype=index_dtype, n_dense=n_dense, n_batch=n_batch), shape=mat.shape)
+
+  @classmethod
+  def from_scipy_sparse(cls, mat, *, index_dtype=None, n_dense=0, n_batch=0):
+    if n_dense != 0 or n_batch != 0:
+      raise NotImplementedError("BCOO.fromscipy with nonzero n_dense/n_batch")
+    mat = mat.tocoo()
+    data = jnp.asarray(mat.data)
+    indices = jnp.vstack([mat.row, mat.col]).astype(index_dtype)
+    return cls((data, indices), shape=mat.shape)
 
   @api.jit
   def todense(self):
@@ -1381,8 +1469,8 @@ class BCOO(JAXSparse):
 
   def tree_flatten(self):
     children = (self.data, self.indices)
-    # pytree sometimes creates dummy objects & we need to handle that.
-    sparse_shape = self.shape if _is_dummy(*children) else self._sparse_shape
+    # pytree sometimes creates placeholder objects & we need to handle that.
+    sparse_shape = self.shape if _is_placeholder(*children) else self._sparse_shape
     # We serialize the sparse shape only to support batching.
     return children, {"sparse_shape": sparse_shape}
 
@@ -1390,8 +1478,8 @@ class BCOO(JAXSparse):
   def tree_unflatten(cls, aux_data, children):
     data, indices = children
     sparse_shape = aux_data["sparse_shape"]
-    # pytree sometimes creates dummy objects & we need to handle that.
-    if _is_dummy(data, indices):
+    # pytree sometimes creates placeholder objects & we need to handle that.
+    if _is_placeholder(data, indices):
       shape = sparse_shape
     else:
       if np.ndim(indices) < 2 or len(sparse_shape) != np.shape(indices)[-2]:
@@ -1403,3 +1491,29 @@ class BCOO(JAXSparse):
           + tuple(sparse_shape)
           + tuple(data.shape[n_batch + 1:]))
     return cls(children, shape=shape)
+
+  # TODO(jakevdp): refactor to avoid circular imports - we can use the same strategy
+  #                we use when adding methods to DeviceArray within lax_numpy.py
+  def __neg__(self):
+    from jax.experimental.sparse import sparsify
+    return sparsify(jnp.negative)(self)
+
+  def __mul__(self, other):
+    from jax.experimental.sparse import sparsify
+    return sparsify(jnp.multiply)(self, other)
+
+  def __rmul__(self, other):
+    from jax.experimental.sparse import sparsify
+    return sparsify(jnp.multiply)(other, self)
+
+  def __add__(self, other):
+    from jax.experimental.sparse import sparsify
+    return sparsify(jnp.add)(self, other)
+
+  def __radd__(self, other):
+    from jax.experimental.sparse import sparsify
+    return sparsify(jnp.add)(other, self)
+
+  def sum(self, *args, **kwargs):
+    from jax.experimental.sparse import sparsify
+    return sparsify(lambda x: x.sum(*args, **kwargs))(self)
