@@ -16,6 +16,7 @@
 from collections import defaultdict, deque
 import itertools as it
 import operator as op
+import re
 from typing import (Any, Callable, Dict, List, Optional, Sequence, Set, Type,
                     Tuple, Union, NamedTuple)
 from warnings import warn
@@ -33,12 +34,12 @@ from jax._src import source_info_util
 from jax._src.abstract_arrays import (make_shaped_array, array_types)
 from ..core import (ConcreteArray, ShapedArray, AbstractToken,
                     Literal, pp_eqn_compact, raise_to_shaped, abstract_token)
+from ..errors import UnexpectedTracerError
 from jax._src.pprint_util import pp
 from .._src.util import (partial, partialmethod, cache, prod, unzip2,
                     extend_name_stack, wrap_name, safe_zip, safe_map)
 from ..lib import xla_bridge as xb
 from ..lib import xla_client as xc
-from ..lib import _xla_extension_version
 from . import partial_eval as pe
 from . import ad
 from . import masking
@@ -52,7 +53,7 @@ xops = xc._xla.ops
 # Types
 Backend = Any  # xc.LocalBackend (why does mypy not like this?)
 Device = Any  # xc.Device
-PyLocalBuffer = Any
+Buffer = Any
 
 XlaOp = Any  # xla_extension.XlaOp
 XlaShape = Any # xla_client.Shape
@@ -62,6 +63,23 @@ XlaExecutable = Any # xla_extension.LocalExecutable
 # This flag is set on exit; no logging should be attempted
 _on_exit = False
 
+
+def compile_or_get_cached(backend, computation, compile_options):
+    # Avoid import cycle between jax and jax.experimental
+    from jax.experimental.compilation_cache import compilation_cache as cc
+    # Persistent compilation cache only implemented on TPU.
+    # TODO(skye): add warning when initializing cache on unsupported default platform
+    if cc.is_initialized() and backend.platform == 'tpu':
+        cached_executable = cc.get_executable(computation, compile_options, backend)
+        if cached_executable is not None:
+            logging.info('Persistent compilation cache hit')
+            return cached_executable
+        else:
+            compiled = backend_compile(backend, computation, compile_options)
+            cc.put_executable(computation, compile_options, compiled, backend)
+            return compiled
+    return backend_compile(backend, computation, compile_options)
+
 def identity(x): return x
 
 _scalar_types = dtypes.python_scalar_dtypes.keys()
@@ -79,6 +97,13 @@ def _make_array_shape(a):
   else:
     return (xc.Shape.array_shape(a.dtype, a.shape),)
 
+def _get_canonical_source_file(frame: source_info_util.Frame):
+  source_file = frame.file_name
+  if config.jax_hlo_source_file_canonicalization_regex:
+    source_file = re.sub(config.jax_hlo_source_file_canonicalization_regex,
+                         '', source_file)
+  return source_file
+
 tracebacks = {}
 def make_op_metadata(primitive: core.Primitive,
                      params: Dict, *,
@@ -90,7 +115,7 @@ def make_op_metadata(primitive: core.Primitive,
   return xc.OpMetadata(
         op_type=primitive.name,
         op_name=str(pp(name_stack) >> pp_eqn_compact(primitive.name, params)),
-        source_file=frame.file_name if frame else None,
+        source_file=_get_canonical_source_file(frame) if frame else None,
         source_line=frame.line_num if frame else None)
 
 ### handlers
@@ -539,12 +564,14 @@ def _axis_groups(mesh_spec, mesh_axes):
       (prod(np.take(mesh_spec, mesh_axes)), -1))
   return tuple(unsafe_map(tuple, groups.T))
 
-def jaxpr_replicas(jaxpr: core.Jaxpr) -> int:
+def jaxpr_replicas(jaxpr) -> int:
   """The number of replicas needed for a jaxpr.
 
   For a eqn, multiply the `axis_size` with the `jaxpr_replicas` of the
   subjaxprs. For a list of eqns, take the maximum number of replicas.
   """
+  if isinstance(jaxpr, core.ClosedJaxpr):
+    jaxpr = jaxpr.jaxpr
   return max(unsafe_map(eqn_replicas, jaxpr.eqns), default=1)
 
 # TODO(mattjj): this function assumes that only pmap has a parameter named
@@ -559,7 +586,7 @@ def eqn_replicas(eqn):
     return 1
 
 def initial_style_primitive_replicas(params):
-  return max(core.traverse_jaxpr_params(jaxpr_replicas, params), default=1)
+  return max(core.traverse_jaxpr_params(jaxpr_replicas, params).values(), default=1)
 
 # TODO(mattjj,skyewm): the functions here are utilities for checking if
 # not-yet-supported features are used with multi-host programming
@@ -592,7 +619,7 @@ def _xla_call_impl(fun: lu.WrappedFun, *args, device, backend, name,
   compiled_fun = _xla_callable(fun, device, backend, name, donated_invars,
                                *unsafe_map(arg_spec, args))
   try:
-    return compiled_fun(*args)
+    out = compiled_fun(*args)
   except FloatingPointError:
     assert config.jax_debug_nans or config.jax_debug_infs  # compiled_fun can only raise in this case
     print("Invalid value encountered in the output of a jit function. "
@@ -603,11 +630,13 @@ def _xla_call_impl(fun: lu.WrappedFun, *args, device, backend, name,
     # but which config.jax_debug_nans is meant to opt into, we'll be re-executing
     # any linear_util.py-style side effects, i.e. re-populating Stores created
     # by any transformation_with_aux's applied to fun. Since this is
-    # intentional here, to avoid "Store occupied" errors we reset the stores to
-    # be empty.
-    for store in fun.stores: store and store.reset()
+    # intentional here, to avoid "Store occupied" errors we clone the WrappedFun
+    # with empty stores.
+    stores = [lu.Store() for _ in fun.stores]
+    clone = lu.WrappedFun(fun.f, fun.transforms, stores, fun.params)
     with core.new_sublevel():
-      return fun.call_wrapped(*args)  # probably won't return
+      _ = clone.call_wrapped(*args)  # probably won't return
+  return out
 
 def flatten_shape(s: XlaShape) -> Sequence[Tuple[Sequence[int], XlaShape]]:
   """Expands a given shape tree into a flat list of indices to arrays.
@@ -667,7 +696,7 @@ def _xla_callable(fun: lu.WrappedFun, device, backend, name, donated_invars, *ar
   jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(
       fun, abstract_args, pe.debug_info_final(fun, "jit"))
   if any(isinstance(c, core.Tracer) for c in consts):
-    raise core.UnexpectedTracerError("Encountered an unexpected tracer.")
+    raise UnexpectedTracerError("Encountered an unexpected tracer.")
   jaxpr, kept_const_idx, kept_var_idx = _prune_unused_inputs(jaxpr)
   consts = [c for i, c in enumerate(consts) if i in kept_const_idx]
   pruned_arg_specs = (a for i, a in enumerate(arg_specs) if i in kept_var_idx)
@@ -740,7 +769,7 @@ def _xla_callable(fun: lu.WrappedFun, device, backend, name, donated_invars, *ar
       num_partitions=1,
       device_assignment=(device.id,) if device else None)
   options.parameter_is_tupled_arguments = tuple_args
-  compiled = backend_compile(backend, built, options)
+  compiled = compile_or_get_cached(backend, built, options)
   if nreps == 1:
     return partial(_execute_compiled, compiled, out_avals, result_handlers,
                    kept_var_idx)
@@ -778,18 +807,8 @@ def set_up_aliases(c, xla_args, out_tuple, donated_args, tuple_args):
   return tuple(out_donated_args)
 
 
-# Pruning unused JIT arguments require jaxlib 0.1.66 or newer.
-# TODO(zhangqiaorjc): remove when jaxlib 0.1.66 is the minimum.
-_ALLOW_ARG_PRUNING = _xla_extension_version >= 20
-
-
 def _prune_unused_inputs(
     jaxpr: core.Jaxpr) -> Tuple[core.Jaxpr, Set[int], Set[int]]:
-  if not _ALLOW_ARG_PRUNING:
-    kept_const_idx = range(len(jaxpr.constvars))
-    kept_var_idx = range(len(jaxpr.invars))
-    return jaxpr, set(kept_const_idx), set(kept_var_idx)
-
   used = {v for v in jaxpr.outvars if isinstance(v, core.Var)}
   # TODO(zhangqiaorjc): Improve the DCE algorithm by also pruning primitive
   # applications that do not produce used outputs. Must handle side-effecting
@@ -1016,7 +1035,8 @@ def lower_fun(fun, multiple_results, parallel=False, with_avals=False, backend=N
     wrapped_fun = lu.wrap_init(fun, params)
     if not multiple_results:
       wrapped_fun = _tuple_output(wrapped_fun)
-    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, avals)
+    with core.extend_axis_env_nd(zip(axis_env.names, axis_env.sizes)):
+      jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, avals)
     outs = jaxpr_subcomp(c, jaxpr, backend, axis_env, _xla_consts(c, consts), '',
                          *xla_args)
     if multiple_results or any(v.aval._num_buffers > 1 for v in jaxpr.outvars):
@@ -1070,8 +1090,8 @@ _CppDeviceArray: DeviceArrayProtocol = xc.Buffer
 def make_device_array(
     aval: core.ShapedArray,
     device: Optional[Device],
-    device_buffer: PyLocalBuffer,
-) -> Union[PyLocalBuffer, "_DeviceArray"]:
+    device_buffer: Buffer,
+) -> Union[Buffer, "_DeviceArray"]:
   """Returns a DeviceArray implementation based on arguments.
 
   This is to be used only within JAX. It will return either a PythonDeviceArray
@@ -1111,11 +1131,9 @@ class _DeviceArray(DeviceArray):  # type: ignore
   """A DeviceArray is an ndarray backed by a single device memory buffer."""
   # We don't subclass ndarray because that would open up a host of issues,
   # but lax_numpy.py overrides isinstance behavior and attaches ndarray methods.
-  # TODO(phawkins): make __weakref__ an unconditional slot when jaxlib 0.1.66
-  # is the minimum version.
   __slots__ = [
-      "aval", "device_buffer", "_npy_value", "_device"
-  ] + ([] if device_array_supports_weakrefs() else ["__weakref__"])
+      "aval", "device_buffer", "_npy_value", "_device", "__weakref__"
+  ]
   __array_priority__ = 100
 
   # DeviceArray has methods that are dynamically populated in lax_numpy.py,
@@ -1123,7 +1141,7 @@ class _DeviceArray(DeviceArray):  # type: ignore
   _HAS_DYNAMIC_ATTRIBUTES = True
 
   def __init__(self, aval: core.ShapedArray, device: Optional[Device],
-               device_buffer: PyLocalBuffer):
+               device_buffer: Buffer):
     """Initializer.
 
     Args:
@@ -1305,14 +1323,12 @@ for device_array in [DeviceArray]:
   setattr(device_array, "__reduce__",
           partialmethod(_forward_to_value, op.methodcaller("__reduce__")))
 
+  # explicitly set to be unhashable.
+  setattr(device_array, "__hash__", None)
+
   # clobbered when jax.numpy is imported, but useful in tests
   setattr(device_array, "__eq__", lambda self, other: self._value == other)
 
-  def __hash__(self):
-    raise TypeError("JAX DeviceArray, like numpy.ndarray, is not hashable.")
-
-  setattr(device_array, "__hash__", __hash__)
-
   # The following methods are dynamically overridden in lax_numpy.py.
   def raise_not_implemented():
     raise NotImplementedError
@@ -1464,14 +1480,18 @@ def _remat_using_while(
 
 def _remat_translation_rule(c, axis_env, in_nodes,
                             name_stack, backend, name, call_jaxpr,
-                            device=None, concrete=None):
+                            prevent_cse, differentiated, concrete, device=None):
   del device, concrete  # Unused.
-  if backend == "gpu":
-    return _remat_using_while(
-        c, axis_env, in_nodes, name_stack, backend, name, call_jaxpr)
+  if differentiated and prevent_cse:
+    if backend == "gpu":
+      return _remat_using_while(
+          c, axis_env, in_nodes, name_stack, backend, name, call_jaxpr)
+    else:
+      return _remat_using_cond(
+          c, axis_env, in_nodes, name_stack, backend, name, call_jaxpr)
   else:
-    return _remat_using_cond(
-        c, axis_env, in_nodes, name_stack, backend, name, call_jaxpr)
+    outs = jaxpr_subcomp(c, call_jaxpr, backend, axis_env, (), "", *in_nodes)
+    return xops.Tuple(c, outs)
 
 call_translations[pe.remat_call_p] = _remat_translation_rule  # type: ignore
 
