@@ -30,7 +30,7 @@ from absl import logging
 logging._warn_preinit_stderr = 0
 
 import jax.lib
-from .._src.config import flags
+from .._src.config import flags, bool_env
 from . import tpu_driver_client
 from . import xla_client
 from jax._src import util, traceback_util
@@ -51,22 +51,26 @@ flags.DEFINE_string(
     'jax_xla_backend is an alias for jax_platform_name. If both are '
     'provided, --jax_xla_backend takes priority. Prefer --jax_platform_name.')
 flags.DEFINE_string(
-    'jax_backend_target', 'local',
+    'jax_backend_target', '',
     'Either "local" or "rpc:address" to connect to a remote service target.')
 flags.DEFINE_string(
     'jax_platform_name',
-    os.getenv('JAX_PLATFORM_NAME', ''),
+    os.getenv('JAX_PLATFORM_NAME', '').lower(),
     'Platform name for XLA. The default is to attempt to use a GPU or TPU if '
     'available, but fall back to CPU otherwise. To set the platform manually, '
-    'pass "cpu" for CPU, "gpu" for GPU, etc.')
+    'pass "cpu" for CPU, "gpu" for GPU, etc. If intending to use CPU, '
+    'setting the platform name to "cpu" can silence warnings that appear with '
+    'the default setting.')
 flags.DEFINE_bool(
-    'jax_disable_most_optimizations', False,
+    'jax_disable_most_optimizations',
+    bool_env('JAX_DISABLE_MOST_OPTIMIZATIONS', False),
     'Try not to do much optimization work. This can be useful if the cost of '
     'optimization is greater than that of running a less-optimized program.')
 flags.DEFINE_string(
-    'jax_cpu_backend_variant', 'tfrt',
-    'jax_cpu_backend_variant selects cpu backend variant: stream_executor or '
-    'tfrt')
+    'jax_cpu_backend_variant',
+     os.getenv('JAX_CPU_BACKEND_VARIANT', 'tfrt'),
+    'Selects CPU backend runtime variant: "stream_executor" or "tfrt". The '
+    'default is "tfrt".')
 
 def get_compile_options(
     num_replicas: int,
@@ -140,6 +144,26 @@ def _make_tpu_driver_client():
   return tpu_driver_client.TpuBackend.create(worker=FLAGS.jax_backend_target)
 
 
+def tpu_client_timer_callback(timer_secs: float):
+  def _log_warning():
+    warnings.warn(
+      f'TPU backend initialization is taking more than {timer_secs} seconds. '
+      'Did you run your code on all TPU hosts? '
+      'See https://jax.readthedocs.io/en/latest/multi_process.html '
+      'for more information.')
+
+  # Will log a warning after `timer_secs`.
+  t = threading.Timer(timer_secs, _log_warning)
+  t.start()
+
+  try:
+    client = xla_client.make_tpu_client()
+  finally:
+    t.cancel()
+
+  return client
+
+
 # Backends, in increasing order of preference.
 # We have no particular opinion about how "backends" relate to "devices". For
 # example, there could be multiple backends that provide the same kind of
@@ -150,48 +174,33 @@ def register_backend_factory(name, factory, *, priority=0):
   _backend_factories[name] = (factory, priority)
 
 
-if jax.lib._xla_extension_version >= 23:
-  register_backend_factory('interpreter', xla_client.make_interpreter_client,
-                           priority=-100)
-  if jax.lib._xla_extension_version >= 24:
-    if FLAGS.jax_cpu_backend_variant == 'stream_executor':
-      register_backend_factory('cpu',
-                               partial(xla_client.make_cpu_client, use_tfrt=False),
-                               priority=0)
-    else:
-      assert FLAGS.jax_cpu_backend_variant == 'tfrt'
-      register_backend_factory('cpu',
-                               partial(xla_client.make_cpu_client, use_tfrt=True),
-                               priority=0)
-  else:
-    register_backend_factory('cpu',
-                              partial(xla_client.make_cpu_client, use_tfrt=False),
-                              priority=0)
-  register_backend_factory('tpu_driver', _make_tpu_driver_client,
-                           priority=100)
-  register_backend_factory('gpu', xla_client.make_gpu_client,
-                           priority=200)
-  register_backend_factory('tpu', xla_client.make_tpu_client,
-                           priority=300)
+register_backend_factory('interpreter', xla_client.make_interpreter_client,
+                         priority=-100)
+if FLAGS.jax_cpu_backend_variant == 'stream_executor':
+  register_backend_factory('cpu',
+                           partial(xla_client.make_cpu_client, use_tfrt=False),
+                           priority=0)
 else:
-  register_backend_factory('interpreter',
-                           xla_client._interpreter_backend_factory,
-                           priority=-100)
-  register_backend_factory('cpu', xla_client._cpu_backend_factory, priority=0)
-  register_backend_factory('tpu_driver', _make_tpu_driver_client,
-                           priority=100)
-  register_backend_factory('gpu', xla_client._gpu_backend_factory,
-                           priority=200)
-  register_backend_factory('tpu', xla_client._tpu_backend_factory,
-                           priority=300)
+  assert FLAGS.jax_cpu_backend_variant == 'tfrt'
+  register_backend_factory('cpu',
+                           partial(xla_client.make_cpu_client, use_tfrt=True),
+                           priority=0)
+register_backend_factory('tpu_driver', _make_tpu_driver_client,
+                         priority=100)
+register_backend_factory('gpu', xla_client.make_gpu_client,
+                         priority=200)
+register_backend_factory(
+  'tpu', partial(tpu_client_timer_callback, timer_secs=60.0), priority=300)
 
 _default_backend = None
 _backends = None
+_backends_errors = None
 _backend_lock = threading.Lock()
 
 
 def backends():
   global _backends
+  global _backends_errors
   global _default_backend
 
   with _backend_lock:
@@ -200,6 +209,7 @@ def backends():
 
     default_priority = -1000
     _backends = {}
+    _backends_errors = {}
     for name, (factory, priority) in _backend_factories.items():
       logging.vlog(1, "Initializing backend '%s'" % name)
       try:
@@ -224,14 +234,14 @@ def backends():
           # If the backend isn't built into the binary, or if it has no devices,
           # we expect a RuntimeError.
           logging.info("Unable to initialize backend '%s': %s" % (name, err))
+          _backends_errors[name] = str(err)
           continue
-    if _default_backend.platform == "cpu":
+    if _default_backend.platform == "cpu" and FLAGS.jax_platform_name != 'cpu':
       logging.warning('No GPU/TPU found, falling back to CPU. '
                       '(Set TF_CPP_MIN_LOG_LEVEL=0 and rerun for more info.)')
     return _backends
 
 
-
 @lru_cache(maxsize=None)  # don't use util.memoize because there is no X64 dependence.
 def get_backend(platform=None):
   # TODO(mattjj,skyewm): remove this input polymorphism after we clean up how
@@ -245,6 +255,9 @@ def get_backend(platform=None):
   if platform is not None:
     backend = bs.get(platform, None)
     if backend is None:
+      if platform in _backends_errors:
+        raise RuntimeError(f"Requested backend {platform}, but it failed "
+                           f"to initialize: {_backends_errors[platform]}")
       raise RuntimeError(f"Unknown backend {platform}")
     return backend
   else:
@@ -254,11 +267,6 @@ def get_backend(platform=None):
 def get_device_backend(device=None):
   """Returns the Backend associated with `device`, or the default Backend."""
   if device is not None:
-    # TODO(phawkins): remove this workaround after jaxlib 0.1.68 becomes the
-    # minimum and it is safe to call `.client` on a tpu_driver TpuDevice.
-    if tpu_driver_client and isinstance(
-        device, tpu_driver_client._tpu_client.TpuDevice):
-      return get_backend('tpu_driver')
     return device.client
   return get_backend()
 
