@@ -13,6 +13,7 @@
 # limitations under the License.
 
 from contextlib import contextmanager
+import inspect
 import functools
 import re
 import os
@@ -37,7 +38,6 @@ from ._src.util import partial, prod, unzip2
 from .tree_util import tree_multimap, tree_all, tree_map, tree_reduce
 from .lib import xla_bridge
 from .interpreters import xla
-from .experimental import maps
 from .experimental.maps import mesh
 
 
@@ -424,16 +424,32 @@ def skip_if_unsupported_type(dtype):
     raise unittest.SkipTest(
       f"Type {dtype.name} not supported on {device_under_test()}")
 
+def is_device_rocm():
+  return xla_bridge.get_backend().platform_version.startswith('rocm')
+
+def is_device_cuda():
+  return xla_bridge.get_backend().platform_version.startswith('cuda')
+
+def _get_device_tags():
+  """returns a set of tags definded for the device under test"""
+  if is_device_rocm():
+    device_tags = set([device_under_test(), "rocm"])
+  elif is_device_cuda():
+    device_tags = set([device_under_test(), "cuda"])
+  else:
+    device_tags = set([device_under_test()])
+  return device_tags
+
 def skip_on_devices(*disabled_devices):
   """A decorator for test methods to skip the test on certain devices."""
   def skip(test_method):
     @functools.wraps(test_method)
     def test_method_wrapper(self, *args, **kwargs):
-      device = device_under_test()
-      if device in disabled_devices:
+      device_tags = _get_device_tags()
+      if device_tags & set(disabled_devices):
         test_name = getattr(test_method, '__name__', '[unknown test]')
         raise unittest.SkipTest(
-          f"{test_name} not supported on {device.upper()}.")
+          f"{test_name} not supported on device with tags {device_tags}.")
       return test_method(self, *args, **kwargs)
     return test_method_wrapper
   return skip
@@ -855,8 +871,18 @@ class JaxTestLoader(absltest.TestLoader):
     return names
 
 
+def with_config(**kwds):
+  """Test case decorator for subclasses of JaxTestCase"""
+  def decorator(cls):
+    assert inspect.isclass(cls) and issubclass(cls, JaxTestCase), "@with_config can only wrap JaxTestCase class definitions."
+    cls._default_config = {**JaxTestCase._default_config, **kwds}
+    return cls
+  return decorator
+
+
 class JaxTestCase(parameterized.TestCase):
   """Base class for JAX tests including numerical checks and boilerplate."""
+  _default_config = {'jax_enable_checks': True}
 
   # TODO(mattjj): this obscures the error messages from failures, figure out how
   # to re-enable it
@@ -864,13 +890,22 @@ class JaxTestCase(parameterized.TestCase):
   #   assert core.reset_trace_state()
 
   def setUp(self):
-    super(JaxTestCase, self).setUp()
-    config.update('jax_enable_checks', True)
+    super().setUp()
+    self._original_config = {}
+    for key, value in self._default_config.items():
+      self._original_config[key] = getattr(config, key)
+      config.update(key, value)
+
     # We use the adler32 hash for two reasons.
     # a) it is deterministic run to run, unlike hash() which is randomized.
     # b) it returns values in int32 range, which RandomState requires.
     self._rng = npr.RandomState(zlib.adler32(self._testMethodName.encode()))
 
+  def tearDown(self):
+    for key, value in self._original_config.items():
+      config.update(key, value)
+    super().tearDown()
+
   def rng(self):
     return self._rng
 
@@ -1048,13 +1083,11 @@ def with_and_without_mesh(f):
 old_spmd_lowering_flag = False
 def set_spmd_lowering_flag(val: bool):
   global old_spmd_lowering_flag
-  maps.make_xmap_callable.cache_clear()
-  old_spmd_lowering_flag = maps.EXPERIMENTAL_SPMD_LOWERING
-  maps.EXPERIMENTAL_SPMD_LOWERING = val
+  old_spmd_lowering_flag = config.experimental_xmap_spmd_lowering
+  config.update('experimental_xmap_spmd_lowering', val)
 
 def restore_spmd_lowering_flag():
-  maps.make_xmap_callable.cache_clear()
-  maps.EXPERIMENTAL_SPMD_LOWERING = old_spmd_lowering_flag
+  config.update('experimental_xmap_spmd_lowering', old_spmd_lowering_flag)
 
 class _cached_property:
   null = object()
